# Proprietary Changes made for Trainy under the Trainy Software License
# Original source: skypilot: https://github.com/skypilot-org/skypilot
# which is Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Amazon Web Services (AWS) S3 Storage."""

import enum
import hashlib
import os
import re
import shlex
import subprocess
import tempfile
import time
import typing
import uuid
from typing import Any, Dict, List, Optional, Tuple

import boto3
import colorama

from konduktor import config, logging
from konduktor.adaptors import aws
from konduktor.data import constants, data_utils, storage_utils
from konduktor.utils import (
    annotations,
    base64_utils,
    common_utils,
    exceptions,
    kubernetes_utils,
    rich_utils,
    ux_utils,
)

logger = logging.get_logger(__name__)

# Maximum number of concurrent rsync upload processes
_MAX_CONCURRENT_UPLOADS = 32

_CREDENTIAL_FILES = ['credentials', 'config']

AWS_SECRET_NAME = 'awscredentials'
AWS_CREDENTIALS_KEY = 'awscredentials'

DEFAULT_AWS_CREDENTIALS_DIR = '~/.aws/'
DEFAULT_AWS_CREDENTIAL_PATH = '~/.aws/credentials'
DEFAULT_AWS_CONFIG_PATH = '~/.aws/config'


class AWSIdentityType(enum.Enum):
    """AWS identity type.

    The account type is determined by the current user identity, based on `aws
    configure list`. We will check the existence of the value in the output of
    `aws configure list` to determine the account type.
    """

    #       Name                    Value             Type    Location
    #       ----                    -----             ----    --------
    #    profile                     1234              env    ...
    # access_key     ****************abcd              sso
    # secret_key     ****************abcd              sso
    #     region                <not set>             None    None
    SSO = 'sso'
    ENV = 'env'
    IAM_ROLE = 'iam-role'
    CONTAINER_ROLE = 'container-role'
    CUSTOM_PROCESS = 'custom-process'
    ASSUME_ROLE = 'assume-role'

    #       Name                    Value             Type    Location
    #       ----                    -----             ----    --------
    #    profile                <not set>             None    None
    # access_key     ****************abcd shared-credentials-file
    # secret_key     ****************abcd shared-credentials-file
    #     region                us-east-1      config-file    ~/.aws/config
    SHARED_CREDENTIALS_FILE = 'shared-credentials-file'

    # IN GCS.PY
    def can_credential_expire(self) -> bool:
        """Check if the AWS identity type can expire.

        SSO,IAM_ROLE and CONTAINER_ROLE are temporary credentials and refreshed
        automatically. ENV and SHARED_CREDENTIALS_FILE are short-lived
        credentials without refresh.
        IAM ROLE:
        https://boto3.amazonaws.com/v1/documentation/api/latest/guide/credentials.html
        SSO/Container-role refresh token:
        https://docs.aws.amazon.com/solutions/latest/dea-api/auth-refreshtoken.html
        """
        # TODO(hong): Add a CLI based check for the expiration of the temporary
        #  credentials
        expirable_types = {AWSIdentityType.ENV, AWSIdentityType.SHARED_CREDENTIALS_FILE}
        return self in expirable_types


class S3Store(storage_utils.AbstractStore):
    """S3Store inherits from Storage Object and represents the backend
    for S3 buckets.
    """

    # k8s secret name for aws credentials
    _AWS_SECRET_NAME = f'{AWS_SECRET_NAME}-{common_utils.user_and_hostname_hash()}'
    _AWS_CREDENTIALS_KEY = AWS_CREDENTIALS_KEY

    _DEFAULT_REGION = 'us-east-1'
    _ACCESS_DENIED_MESSAGE = 'Access Denied'
    _CUSTOM_ENDPOINT_REGIONS = [
        'ap-east-1',
        'me-south-1',
        'af-south-1',
        'eu-south-1',
        'eu-south-2',
        'ap-south-2',
        'ap-southeast-3',
        'ap-southeast-4',
        'me-central-1',
        'il-central-1',
    ]

    _INDENT_PREFIX = '    '

    _STATIC_CREDENTIAL_HELP_STR = (
        'Run the following commands:'
        f'\n{_INDENT_PREFIX}  $ aws configure'
        f'\n{_INDENT_PREFIX}  $ aws configure list '
        '# Ensure that this shows identity is set.'
        f'\n{_INDENT_PREFIX}For more info: '
        'https://docs.aws.amazon.com/cli/latest/userguide/cli-configure-quickstart.html'  # pylint: disable=line-too-long
    )

    _REPR = 'S3Store'

    def __init__(
        self,
        name: str,
        source: str,
        region: Optional[str] = _DEFAULT_REGION,
        is_sky_managed: Optional[bool] = False,
        sync_on_reconstruction: Optional[bool] = True,
        _bucket_sub_path: Optional[str] = None,
    ):
        self.client: 'boto3.client.Client'
        self.bucket: 'constants.StorageHandle'
        if region in self._CUSTOM_ENDPOINT_REGIONS:
            logger.warning(
                'AWS opt-in regions are not supported for S3. '
                f'Falling back to default region '
                f'{self._DEFAULT_REGION} for bucket {name!r}.'
            )
            region = self._DEFAULT_REGION
        super().__init__(
            name,
            source,
            region,
            is_sky_managed,
            sync_on_reconstruction,
            _bucket_sub_path,
        )

    def __repr__(self):
        return self._REPR

    # IN GCS.PY
    def _validate(self):
        if self.source is not None and isinstance(self.source, str):
            if self.source.startswith('s3://'):
                assert self.name == data_utils.split_s3_path(self.source)[0], (
                    'S3 Bucket is specified as path, the name should be the'
                    ' same as S3 bucket.'
                )
                assert data_utils.verify_s3_bucket(self.name), (
                    f'Source specified as {self.source}, an S3 bucket. ',
                    'S3 Bucket should exist.',
                )
            # if self.source.startswith('gs://'):
            #     assert self.name == data_utils.split_gcs_path(self.source)[0], (
            #         'GCS Bucket is specified as path, the name should be '
            #         'the same as GCS bucket.'
            #     )
            # elif data_utils.is_az_container_endpoint(self.source):
            #     storage_account_name, container_name, _ = (
            #         data_utils.split_az_path(self.source))
            #     assert self.name == container_name, (
            #         'Azure bucket is specified as path, the name should be '
            #         'the same as Azure bucket.')
            #     assert data_utils.verify_az_bucket(
            #         storage_account_name, self.name), (
            #             f'Source specified as {self.source}, an Azure bucket. '
            #             'Azure bucket should exist.')
            # elif self.source.startswith('r2://'):
            #     assert self.name == data_utils.split_r2_path(self.source)[0], (
            #         'R2 Bucket is specified as path, the name should be '
            #         'the same as R2 bucket.')
            #     assert data_utils.verify_r2_bucket(self.name), (
            #         f'Source specified as {self.source}, a R2 bucket. ',
            #         'R2 Bucket should exist.')
            # elif self.source.startswith('cos://'):
            #     assert self.name == data_utils.split_cos_path(self.source)[0], (
            #         'COS Bucket is specified as path, the name should be '
            #         'the same as COS bucket.')
            #     assert data_utils.verify_ibm_cos_bucket(self.name), (
            #         f'Source specified as {self.source}, a COS bucket. ',
            #         'COS Bucket should exist.')
        # Validate name
        self.name = self.validate_name(self.name)

    # IN GCS.PY
    @classmethod
    def validate_name(cls, name: str) -> str:
        """Validates the name of the S3 store.

        Source for rules:
        https://docs.aws.amazon.com/AmazonS3/latest/userguide/bucketnamingrules.html
        """

        def _raise_no_traceback_name_error(err_str):
            with ux_utils.print_exception_no_traceback():
                raise exceptions.StorageNameError(err_str)

        if name is not None and isinstance(name, str):
            # Check for overall length
            if not 3 <= len(name) <= 63:
                _raise_no_traceback_name_error(
                    f'Invalid store name: name {name} must be between 3 (min) '
                    'and 63 (max) characters long.'
                )

            # Check for valid characters and start/end with a number or letter
            pattern = r'^[a-z0-9][-a-z0-9._]*[a-z0-9]$'
            if not re.match(pattern, name):
                _raise_no_traceback_name_error(
                    f'Invalid store name: name {name} can consist only of '
                    'lowercase letters, numbers, dots (.), and hyphens (-). '
                    'It must begin and end with a letter or number.'
                )

            # Check for two adjacent periods
            if '..' in name:
                _raise_no_traceback_name_error(
                    f'Invalid store name: name {name} must not contain '
                    'two adjacent periods.'
                )

            # Check for IP address format
            ip_pattern = r'^(?:\d{1,3}\.){3}\d{1,3}$'
            if re.match(ip_pattern, name):
                _raise_no_traceback_name_error(
                    f'Invalid store name: name {name} must not be formatted as '
                    'an IP address (for example, 192.168.5.4).'
                )

            # Check for 'xn--' prefix
            if name.startswith('xn--'):
                _raise_no_traceback_name_error(
                    f'Invalid store name: name {name} must not start with the '
                    'prefix "xn--".'
                )

            # Check for '-s3alias' suffix
            if name.endswith('-s3alias'):
                _raise_no_traceback_name_error(
                    f'Invalid store name: name {name} must not end with the '
                    'suffix "-s3alias".'
                )

            # Check for '--ol-s3' suffix
            if name.endswith('--ol-s3'):
                _raise_no_traceback_name_error(
                    f'Invalid store name: name {name} must not end with the '
                    'suffix "--ol-s3".'
                )
        else:
            _raise_no_traceback_name_error('Store name must be specified.')
        return name

    # IN GCS.PY
    def initialize(self):
        """Initializes the S3 store object on the cloud.

        Initialization involves fetching bucket if exists, or creating it if
        it does not.

        Raises:
          StorageBucketCreateError: If bucket creation fails
          StorageBucketGetError: If fetching existing bucket fails
          StorageInitError: If general initialization fails.
        """
        self.client = data_utils.create_s3_client(self.region)
        self.bucket, is_new_bucket = self._get_bucket()
        if self.is_sky_managed is None:
            # If is_sky_managed is not specified, then this is a new storage
            # object (i.e., did not exist in global_user_state) and we should
            # set the is_sky_managed property.
            # If is_sky_managed is specified, then we take no action.
            self.is_sky_managed = is_new_bucket

    # IN GCS.PY
    def upload(self):
        """Uploads source to store bucket.

        Upload must be called by the Storage handler - it is not called on
        Store initialization.

        Raises:
            StorageUploadError: if upload fails.
        """
        try:
            if isinstance(self.source, list):
                self.batch_aws_rsync(self.source, create_dirs=True)
            elif self.source is not None:
                if self.source.startswith('s3://'):
                    pass
                # elif self.source.startswith('gs://'):
                #     self._transfer_to_s3()
                # elif self.source.startswith('r2://'):
                #     self._transfer_to_s3()
                else:
                    self.batch_aws_rsync([self.source])
        except exceptions.StorageUploadError:
            raise
        except Exception as e:
            raise exceptions.StorageUploadError(
                f'Upload failed for store {self.name}'
            ) from e

    # IN GCS.PY
    def delete(self) -> None:
        deleted_by_skypilot = self._delete_s3_bucket(self.name)
        if deleted_by_skypilot:
            msg_str = f'Deleted S3 bucket {self.name}.'
        else:
            msg_str = (
                f'S3 bucket {self.name} may have been deleted '
                f'externally. Removing from local state.'
            )
        logger.info(f'{colorama.Fore.GREEN}{msg_str}' f'{colorama.Style.RESET_ALL}')

    # IN GCS.PY
    def get_handle(self) -> 'constants.StorageHandle':
        return aws.resource('s3').Bucket(self.name)

    # FROM data/storage.py but matches GCS.PY batch_gsutil_rsync() (s3 specific)
    def batch_aws_rsync(
        self, source_path_list: List['constants.Path'], create_dirs: bool = False
    ) -> None:
        """Invokes aws s3 sync to batch upload a list of local paths to S3

        AWS Sync by default uses 10 threads to upload files to the bucket.  To
        increase parallelism, modify max_concurrent_requests in your aws config
        file (Default path: ~/.aws/config).

        Since aws s3 sync does not support batch operations, we construct
        multiple commands to be run in parallel.

        Args:
            source_path_list: List of paths to local files or directories
            create_dirs: If the local_path is a directory and this is set to
                False, the contents of the directory are directly uploaded to
                root of the bucket. If the local_path is a directory and this is
                set to True, the directory is created in the bucket root and
                contents are uploaded to it.
        """
        sub_path = f'/{self._bucket_sub_path}' if self._bucket_sub_path else ''

        def get_file_sync_command(base_dir_path, file_names):
            includes = ' '.join(
                [f'--include {shlex.quote(file_name)}' for file_name in file_names]
            )
            base_dir_path = shlex.quote(base_dir_path)
            sync_command = (
                'aws s3 sync --no-follow-symlinks --exclude="*" '
                f'{includes} {base_dir_path} '
                f's3://{self.name}{sub_path}'
            )
            return sync_command

        def get_dir_sync_command(src_dir_path, dest_dir_name):
            # we exclude .git directory from the sync
            excluded_list = storage_utils.get_excluded_files(src_dir_path)
            excluded_list.append('.git/*')
            excludes = ' '.join(
                [f'--exclude {shlex.quote(file_name)}' for file_name in excluded_list]
            )
            src_dir_path = shlex.quote(src_dir_path)
            sync_command = (
                f'aws s3 sync --no-follow-symlinks {excludes} '
                f'{src_dir_path} '
                f's3://{self.name}{sub_path}/{dest_dir_name}'
            )
            return sync_command

        # Generate message for upload
        if len(source_path_list) > 1:
            source_message = f'{len(source_path_list)} paths'
        else:
            source_message = source_path_list[0]

        log_path = logging.generate_tmp_logging_file_path(
            constants._STORAGE_LOG_FILE_NAME
        )
        sync_path = f'{source_message} -> s3://{self.name}{sub_path}/'
        with rich_utils.safe_status(
            ux_utils.spinner_message(f'Syncing {sync_path}', log_path=log_path)
        ):
            data_utils.parallel_upload(
                source_path_list,
                get_file_sync_command,
                get_dir_sync_command,
                log_path,
                self.name,
                self._ACCESS_DENIED_MESSAGE,
                create_dirs=create_dirs,
                max_concurrent_uploads=_MAX_CONCURRENT_UPLOADS,
            )
        logger.info(
            ux_utils.finishing_message(f'Storage synced: {sync_path}', log_path)
        )

    # IN GCS.PY
    def _get_bucket(self) -> Tuple['constants.StorageHandle', bool]:
        """Obtains the S3 bucket.

        If the bucket exists, this method will return the bucket.
        If the bucket does not exist, there are three cases:
          1) Raise an error if the bucket source starts with s3://
          2) Return None if bucket has been externally deleted and
             sync_on_reconstruction is False
          3) Create and return a new bucket otherwise

        Raises:
            StorageSpecError: If externally created bucket is attempted to be
                mounted without specifying storage source.
            StorageBucketCreateError: If creating the bucket fails
            StorageBucketGetError: If fetching a bucket fails
            StorageExternalDeletionError: If externally deleted storage is
                attempted to be fetched while reconstructing the storage for
                'sky storage delete' or 'sky start'
        """
        s3 = aws.resource('s3')
        bucket = s3.Bucket(self.name)

        try:
            # Try Public bucket case.
            # This line does not error out if the bucket is an external public
            # bucket or if it is a user's bucket that is publicly
            # accessible.
            self.client.head_bucket(Bucket=self.name)
            self._validate_existing_bucket()
            return bucket, False
        except aws.botocore_exceptions().ClientError as e:
            error_code = e.response['Error']['Code']
            # AccessDenied error for buckets that are private and not owned by
            # user.
            if error_code == '403':
                command = f'aws s3 ls {self.name}'
                with ux_utils.print_exception_no_traceback():
                    raise exceptions.StorageBucketGetError(
                        f'Bucket {self.name} does not exist.'
                        + f' To debug, consider running `{command}`.'
                    ) from e

        if isinstance(self.source, str) and self.source.startswith('s3://'):
            with ux_utils.print_exception_no_traceback():
                raise exceptions.StorageBucketGetError(
                    'Attempted to use a non-existent bucket as a source: '
                    f'{self.source}. Consider using `aws s3 ls '
                    f'{self.source}` to debug.'
                )

        # If bucket cannot be found in both private and public settings,
        # the bucket is to be created by Sky. However, creation is skipped if
        # Store object is being reconstructed for deletion or re-mount with
        # sky start, and error is raised instead.
        if self.sync_on_reconstruction:
            bucket = self._create_s3_bucket(self.name, self.region)
            return bucket, True
        else:
            # Raised when Storage object is reconstructed for sky storage
            # delete or to re-mount Storages with sky start but the storage
            # is already removed externally.
            raise exceptions.StorageExternalDeletionError(
                'Attempted to fetch a non-existent bucket: ' f'{self.name}'
            )

    # IN GCS.PY
    def _download_file(self, remote_path: str, local_path: str) -> None:
        """Downloads file from remote to local on s3 bucket
        using the boto3 API

        Args:
          remote_path: str; Remote path on S3 bucket
          local_path: str; Local path on user's device
        """
        self.bucket.download_file(remote_path, local_path)

    # IN GCS.PY
    def _create_s3_bucket(
        self, bucket_name: str, region=_DEFAULT_REGION
    ) -> 'constants.StorageHandle':
        """Creates S3 bucket with specific name in specific region

        Args:
          bucket_name: str; Name of bucket
          region: str; Region name, e.g. us-west-1, us-east-2
        Raises:
          StorageBucketCreateError: If bucket creation fails.
        """
        s3_client = self.client
        try:
            create_bucket_config: Dict[str, Any] = {'Bucket': bucket_name}
            # If default us-east-1 region of create_bucket API is used,
            # the LocationConstraint must not be specified.
            # Reference: https://stackoverflow.com/a/51912090
            if region is not None and region != 'us-east-1':
                create_bucket_config['CreateBucketConfiguration'] = {
                    'LocationConstraint': region
                }
            s3_client.create_bucket(**create_bucket_config)
            logger.info(
                f'  {colorama.Style.DIM}Created S3 bucket {bucket_name!r} in '
                f'{region or "us-east-1"}{colorama.Style.RESET_ALL}'
            )

            # Add AWS tags configured in config.yaml to the bucket.
            # This is useful for cost tracking and external cleanup.
            bucket_tags = config.get_nested(('aws', 'labels'), {})
            if bucket_tags:
                s3_client.put_bucket_tagging(
                    Bucket=bucket_name,
                    Tagging={
                        'TagSet': [
                            {'Key': k, 'Value': v} for k, v in bucket_tags.items()
                        ]
                    },
                )

        except aws.botocore_exceptions().ClientError as e:
            with ux_utils.print_exception_no_traceback():
                raise exceptions.StorageBucketCreateError(
                    f'Attempted to create a bucket {self.name} but failed.'
                ) from e
        return aws.resource('s3').Bucket(bucket_name)

    # NOT IN GCS.PY but FROM data/storage.py (s3 specific)
    def _execute_s3_remove_command(
        self, command: str, bucket_name: str, hint_operating: str, hint_failed: str
    ) -> bool:
        try:
            with rich_utils.safe_status(ux_utils.spinner_message(hint_operating)):
                subprocess.check_output(command.split(' '), stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            if 'NoSuchBucket' in e.output.decode('utf-8'):
                logger.debug(f'Bucket {bucket_name} does not exist.')
                return False
            else:
                with ux_utils.print_exception_no_traceback():
                    raise exceptions.StorageBucketDeleteError(
                        f'{hint_failed}' f'Detailed error: {e.output}'
                    )
        return True

    # IN GCS.PY
    def _delete_s3_bucket(self, bucket_name: str) -> bool:
        """Deletes S3 bucket, including all objects in bucket

        Args:
            bucket_name: str; Name of bucket

        Returns:
            bool; True if bucket was deleted, False if it was deleted externally.

        Raises:
            StorageBucketDeleteError: If deleting the bucket fails.
        """
        # Deleting objects is very slow programatically
        # (i.e. bucket.objects.all().delete() is slow).
        # In addition, standard delete operations (i.e. via `aws s3 rm`)
        # are slow, since AWS puts deletion markers.
        # https://stackoverflow.com/questions/49239351/why-is-it-so-much-slower-to-delete-objects-in-aws-s3-than-it-is-to-create-them
        # The fastest way to delete is to run `aws s3 rb --force`,
        # which removes the bucket by force.
        remove_command = f'aws s3 rb s3://{bucket_name} --force'
        success = self._execute_s3_remove_command(
            remove_command,
            bucket_name,
            f'Deleting S3 bucket [green]{bucket_name}[/]',
            f'Failed to delete S3 bucket {bucket_name}.',
        )
        if not success:
            return False

        # Wait until bucket deletion propagates on AWS servers
        while data_utils.verify_s3_bucket(bucket_name):
            time.sleep(0.1)
        return True

    # NOT IN GCS.PY but FROM data/storage.py (s3 specific)
    def _delete_s3_bucket_sub_path(self, bucket_name: str, sub_path: str) -> bool:
        """Deletes the sub path from the bucket."""
        remove_command = f'aws s3 rm s3://{bucket_name}/{sub_path}/ --recursive'
        return self._execute_s3_remove_command(
            remove_command,
            bucket_name,
            f'Removing objects from S3 bucket ' f'[green]{bucket_name}/{sub_path}[/]',
            f'Failed to remove objects from S3 bucket {bucket_name}/{sub_path}.',
        )

    @classmethod
    @annotations.lru_cache(scope='global', maxsize=1)
    def _aws_configure_list(cls) -> Optional[bytes]:
        proc = subprocess.run(
            'aws configure list',
            shell=True,
            check=False,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        if proc.returncode != 0:
            return None
        return proc.stdout

    @classmethod
    def _sso_credentials_help_str(cls, expired: bool = False) -> str:
        help_str = 'Run the following commands (must use AWS CLI v2):'
        if not expired:
            help_str += f'\n{cls._INDENT_PREFIX}  $ aws configure sso'
        help_str += (
            f'\n{cls._INDENT_PREFIX}  $ aws sso login --profile <profile_name>'
            f'\n{cls._INDENT_PREFIX}For more info: '
            'https://docs.aws.amazon.com/cli/latest/userguide/cli-configure-sso.html'
        )
        return help_str

    @classmethod
    @annotations.lru_cache(
        scope='global', maxsize=1
    )  # Cache since getting identity is slow.
    def _sts_get_caller_identity(cls) -> Optional[List[List[str]]]:
        try:
            sts = aws.client('sts', check_credentials=False)
            # The caller identity contains 3 fields: UserId, Account, Arn.
            # 1. 'UserId' is unique across all AWS entity, which looks like
            # "AROADBQP57FF2AEXAMPLE:role-session-name"
            # 2. 'Account' can be shared by multiple users under the same
            # organization
            # 3. 'Arn' is the full path to the user, which can be reused when
            # the user is deleted and recreated.
            # Refer to: <https://docs.aws.amazon.com/cli/latest/reference/sts/get-caller-identity.html>
            # and <https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_policies_variables.html#principaltable>
            user_info = sts.get_caller_identity()
            # Allow fallback to AccountId if UserId does not match, because:
            # 1. In the case where multiple IAM users belong a single root account,
            # those users normally share the visibility of the VMs, so we do not
            # need to identity them with each other. (There can be some cases,
            # when an IAM user is given a limited permission by the admin, we may
            # ignore that case for now, or print out a warning if the underlying
            # userid changed for a cluster).
            # 2. In the case where the multiple users belong to an organization,
            # those users will have different account id, so fallback works.
            user_ids = [user_info['UserId'], user_info['Account']]
        except aws.botocore_exceptions().NoCredentialsError as e:
            with ux_utils.print_exception_no_traceback():
                raise exceptions.CloudUserIdentityError(
                    'AWS credentials are not set. '
                    f'{cls._STATIC_CREDENTIAL_HELP_STR}\n'
                    f'{cls._INDENT_PREFIX}Details: `aws sts '
                    'get-caller-identity` failed with error:'
                    f' {common_utils.format_exception(e, use_bracket=True)}.'
                ) from None
        except aws.botocore_exceptions().ClientError as e:
            with ux_utils.print_exception_no_traceback():
                raise exceptions.CloudUserIdentityError(
                    'Failed to access AWS services with credentials. '
                    'Make sure that the access and secret keys are correct.'
                    f' {cls._STATIC_CREDENTIAL_HELP_STR}\n'
                    f'{cls._INDENT_PREFIX}Details: `aws sts '
                    'get-caller-identity` failed with error:'
                    f' {common_utils.format_exception(e, use_bracket=True)}.'
                ) from None
        except aws.botocore_exceptions().InvalidConfigError as e:
            # pylint: disable=import-outside-toplevel
            import awscli
            from packaging import version

            awscli_version = version.parse(awscli.__version__)
            if awscli_version < version.parse(
                '1.27.10'
            ) and 'configured to use SSO' in str(e):
                with ux_utils.print_exception_no_traceback():
                    raise exceptions.CloudUserIdentityError(
                        'awscli is too old to use SSO.'
                        'Run the following command to upgrade:'
                        f'\n{cls._INDENT_PREFIX}  $ pip install awscli>=1.27.10'
                        f'\n{cls._INDENT_PREFIX}You may need to log into SSO again '
                        f'after upgrading. {cls._sso_credentials_help_str()}'
                    ) from None
            with ux_utils.print_exception_no_traceback():
                raise exceptions.CloudUserIdentityError(
                    f'Invalid AWS configuration.\n'
                    f'  Reason: {common_utils.format_exception(e, use_bracket=True)}.'
                ) from None
        except aws.botocore_exceptions().TokenRetrievalError:
            # This is raised when the access token is expired, which mainly
            # happens when the user is using temporary credentials or SSO
            # login.
            with ux_utils.print_exception_no_traceback():
                raise exceptions.CloudUserIdentityError(
                    'AWS access token is expired.'
                    f' {cls._sso_credentials_help_str(expired=True)}'
                ) from None
        except Exception as e:  # pylint: disable=broad-except
            with ux_utils.print_exception_no_traceback():
                raise exceptions.CloudUserIdentityError(
                    f'Failed to get AWS user.\n'
                    f'  Reason: {common_utils.format_exception(e, use_bracket=True)}.'
                ) from None
        # TODO: Return a list of identities in the profile when we support
        #   automatic switching for AWS. Currently we only support one identity.
        return [user_ids]

    # IN GCS.PY
    @classmethod
    @annotations.lru_cache(
        scope='global', maxsize=1
    )  # Cache since getting identity is slow.
    def get_user_identities(cls) -> List[List[str]]:
        """Returns a [UserId, Account] list that uniquely identifies the user.

        These fields come from `aws sts get-caller-identity` and are cached
        locally by `aws configure list` output. The identities are assumed to
        be stable for the duration of the `sky` process. Modifying the
        credentials while the `sky` process is running will not affect the
        identity returned by this function.

        We permit the same actual user to:

          - switch between different root accounts (after which both elements
            of the list will be different) and have their clusters owned by
            each account be protected; or

          - within the same root account, switch between different IAM
            users, and treat [user_id=1234, account=A] and
            [user_id=4567, account=A] to be the *same*. Namely, switching
            between these IAM roles within the same root account will cause
            the first element of the returned list to differ, and will allow
            the same actual user to continue to interact with their clusters.
            Note: this is not 100% safe, since the IAM users can have very
            specific permissions, that disallow them to access the clusters
            but it is a reasonable compromise as that could be rare.

        Returns:
            A list of strings that uniquely identifies the user on this cloud.
            For identity check, we will fallback through the list of strings
            until we find a match, and print a warning if we fail for the
            first string.

        Raises:
            exceptions.CloudUserIdentityError: if the user identity cannot be
                retrieved.
        """
        stdout = cls._aws_configure_list()
        if stdout is None:
            # `aws configure list` is not available, possible reasons:
            # - awscli is not installed but credentials are valid, e.g. run from
            #   an EC2 instance with IAM role
            # - aws credentials are not set, proceed anyway to get unified error
            #   message for users
            return cls._sts_get_caller_identity()
        config_hash = hashlib.md5(stdout).hexdigest()[:8]  # noqa: F841
        # Getting aws identity cost ~1s, so we cache the result with the output of
        # `aws configure list` as cache key. Different `aws configure list` output
        # can have same aws identity, our assumption is the output would be stable
        # in real world, so the number of cache files would be limited.
        # TODO(aylei): consider using a more stable cache key and evalute eviction.
        # TODO:(ryan) ??? Ignoring caching for now (returning early)
        return cls._sts_get_caller_identity()
        # cache_path = catalog_common.get_catalog_path(
        #     f'aws/.cache/user-identity-{config_hash}.txt')
        # if os.path.exists(cache_path):
        #     try:
        #         with open(cache_path, 'r', encoding='utf-8') as f:
        #             return json.loads(f.read())
        #     except json.JSONDecodeError:
        #         # cache is invalid, ignore it and fetch identity again
        #         pass
        #
        # result = cls._sts_get_caller_identity()
        # with open(cache_path, 'w', encoding='utf-8') as f:
        #     f.write(json.dumps(result))
        # return result

    # IN GCS.PY
    @classmethod
    def get_active_user_identity_str(cls) -> Optional[str]:
        user_identity = cls.get_active_user_identity()
        if user_identity is None:
            return None
        identity_str = f'{user_identity[0]} [account={user_identity[1]}]'
        return identity_str

    # IN GCS.PY
    @classmethod
    @annotations.lru_cache(scope='global', maxsize=1)
    def check_credentials(cls) -> Tuple[bool, Optional[str]]:
        """Checks if the user has access credentials to AWS."""

        dependency_installation_hints = (
            'AWS dependencies are not installed. '
            'Run the following commands:'
            f'\n{cls._INDENT_PREFIX}  $ pip install boto3 botocore awscli'
            f'\n{cls._INDENT_PREFIX}Credentials may also need to be set. '
            f'{cls._STATIC_CREDENTIAL_HELP_STR}'
        )

        # Checks if the AWS CLI is installed properly
        proc = subprocess.run(
            'aws --version',
            shell=True,
            check=False,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        if proc.returncode != 0:
            return False, dependency_installation_hints
        try:
            # Checks if aws boto is installed properly
            # pylint: disable=import-outside-toplevel, unused-import
            import boto3  # noqa: F401
            import botocore  # noqa: F401
        except ImportError:
            return False, dependency_installation_hints

        # Checks if AWS credentials 1) exist and 2) are valid.
        # https://stackoverflow.com/questions/53548737/verify-aws-credentials-with-boto3
        try:
            identity_str = cls.get_active_user_identity_str()  # noqa: F841
        except exceptions.CloudUserIdentityError as e:
            return False, str(e)

        static_credential_exists = os.path.isfile(
            os.path.expanduser('~/.aws/credentials')
        )
        hints = None
        identity_type = cls._current_identity_type()
        single_cloud_hint = (
            ' It will work if you use AWS only, but will cause problems '
            'if you want to use multiple clouds. To set up static credentials, '
            'try: aws configure'
        )
        if identity_type == AWSIdentityType.SSO:
            hints = 'AWS SSO is set.'
            if static_credential_exists:
                hints += (
                    ' To ensure multiple clouds work correctly, please use SkyPilot '
                    'with static credentials (e.g., ~/.aws/credentials) by unsetting '
                    'the AWS_PROFILE environment variable.'
                )
            else:
                hints += single_cloud_hint
        elif identity_type == AWSIdentityType.IAM_ROLE:
            # When using an IAM role, the credentials may not exist in the
            # ~/.aws/credentials file. So we don't check for the existence of the
            # file. This will happen when the user is on a VM (or
            # jobs-controller) created by an SSO account, i.e. the VM will be
            # assigned the IAM role: skypilot-v1.
            hints = f'AWS IAM role is set.{single_cloud_hint}'
        elif identity_type == AWSIdentityType.CONTAINER_ROLE:
            # Similar to the IAM ROLE, an ECS container may not store credentials
            # in the ~/.aws/credentials file. So we don't check for the existence of
            # the file. i.e. the container will be assigned the IAM role of the
            # task: skypilot-v1.
            hints = f'AWS container-role is set.{single_cloud_hint}'
        elif identity_type == AWSIdentityType.CUSTOM_PROCESS:
            # Similar to the IAM ROLE, a custom process may not store credentials
            # in the ~/.aws/credentials file. So we don't check for the existence of
            # the file. i.e. the custom process will be assigned the IAM role of the
            # task: skypilot-v1.
            hints = f'AWS custom-process is set.{single_cloud_hint}'
        elif identity_type == AWSIdentityType.ASSUME_ROLE:
            # When using ASSUME ROLE, the credentials are coming from a different
            # source profile. So we don't check for the existence of ~/.aws/credentials.
            # i.e. the assumed role will be assigned the IAM role of the
            # task: skypilot-v1.
            hints = f'AWS assume-role is set.{single_cloud_hint}'
        elif identity_type == AWSIdentityType.ENV:
            # When using ENV vars, the credentials are coming from the environment
            # variables. So we don't check for the existence of ~/.aws/credentials.
            # i.e. the identity is not determined by the file.
            hints = f'AWS env is set.{single_cloud_hint}'
        else:
            # This file is required because it is required by the VMs launched on
            # other clouds to access private s3 buckets and resources like EC2.
            # `get_active_user_identity` does not guarantee this file exists.
            if not static_credential_exists:
                return (
                    False,
                    '~/.aws/credentials does not exist. '
                    + cls._STATIC_CREDENTIAL_HELP_STR,
                )

        try:
            s3 = boto3.client('s3')

            suffix = uuid.uuid4().hex[:6]
            test_bucket = f'konduktor-check-s3-{int(time.time())}-{suffix}'

            try:
                s3.create_bucket(Bucket=test_bucket)

                time.sleep(1)

                s3.get_bucket_location(Bucket=test_bucket)
                s3.list_objects_v2(Bucket=test_bucket, MaxKeys=1)

                # Object-related checks
                s3.put_object(Bucket=test_bucket, Key='test.txt', Body='hello')
                s3.get_object(Bucket=test_bucket, Key='test.txt')
                s3.delete_object(Bucket=test_bucket, Key='test.txt')

            finally:
                # Always attempt to clean up, even if earlier checks failed
                try:
                    s3.delete_bucket(Bucket=test_bucket)
                except Exception:
                    raise RuntimeError(
                        'AWS S3 delete bucket permission is missing. '
                        'Please check your policy.\n'
                    )

        except Exception:
            return False, (
                'One or more AWS S3 permissions are missing. '
                'Please check your policy.\n'
            )

        logger.info(
            f'AWS credentials are valid '
            f'for the current identity {logging.CHECK_MARK_EMOJI}'
        )
        logger.info('Creating k8s secret with AWS credentials...')
        set_ok, result = cls.set_secret_credentials()
        if not set_ok:
            logger.error(f'Failed to create k8s secret with AWS credentials: {result}')
            return False, result
        return True, hints

    @classmethod
    def _current_identity_type(cls) -> Optional[AWSIdentityType]:
        stdout = cls._aws_configure_list()
        if stdout is None:
            return None
        output = stdout.decode()

        # We determine the identity type by looking at the output of
        # `aws configure list`. The output looks like:
        #   Name                   Value         Type    Location
        #   ----                   -----         ----    --------
        #   profile                <not set>     None    None
        #   access_key     *       <not set>     sso     None
        #   secret_key     *       <not set>     sso     None
        #   region                 <not set>     None    None
        # We try to determine the identity type by looking for the
        # string "sso"/"iam-role" in the output, i.e. the "Type" column.

        def _is_access_key_of_type(type_str: str) -> bool:
            # The dot (.) does not match line separators.
            results = re.findall(rf'access_key.*{type_str}', output)
            if len(results) > 1:
                raise RuntimeError(f'Unexpected `aws configure list` output:\n{output}')
            return len(results) == 1

        for identity_type in AWSIdentityType:
            if _is_access_key_of_type(identity_type.value):
                return identity_type
        return AWSIdentityType.SHARED_CREDENTIALS_FILE

    # IN GCS.PY
    @classmethod
    def set_secret_credentials(cls) -> Tuple[bool, Optional[str]]:
        """
        Set the k8s secret storing the AWS credentials
        """
        context = kubernetes_utils.get_current_kube_config_context_name()
        namespace = kubernetes_utils.get_kube_config_context_namespace()

        # Check if credentials are provided via environment
        access_key = os.environ.get('AWS_ACCESS_KEY_ID')
        secret_key = os.environ.get('AWS_SECRET_ACCESS_KEY')
        region = os.environ.get('AWS_DEFAULT_REGION', 'us-east-1')

        if access_key and secret_key:
            logger.info('Using AWS credentials from env')
            credentials_dir = tempfile.mkdtemp()
            credentials_path = os.path.join(credentials_dir, 'credentials')
            config_path = os.path.join(credentials_dir, 'config')

            with open(credentials_path, 'w') as f:
                f.write(f"""[default]
            aws_access_key_id = {access_key}
            aws_secret_access_key = {secret_key}
            """)

            with open(config_path, 'w') as f:
                f.write(f"""[default]
            region = {region}
            """)
        else:
            logger.info('Using AWS credentils from ~/.aws/')
            credentials_dir = DEFAULT_AWS_CREDENTIALS_DIR

        credentials_files = [
            os.path.expanduser(os.path.join(credentials_dir, f))
            for f in _CREDENTIAL_FILES
        ]
        ok, result = kubernetes_utils.set_secret(
            secret_name=cls._AWS_SECRET_NAME,
            namespace=namespace,
            context=context,
            secret_key=cls._AWS_CREDENTIALS_KEY,
            secret_value=base64_utils.zip_base64encode(credentials_files),
        )
        if not ok:
            logger.error(f'Failed to set AWS credentials in k8s secret: \n{result}')
            return False, result
        else:
            logger.info(
                f'AWS credentials set in k8s secret: {cls._AWS_SECRET_NAME} '
                f'in namespace {namespace} in context {context} '
                f'{logging.CHECK_MARK_EMOJI}'
            )

        try:
            identity = boto3.client('sts').get_caller_identity()
            logger.info(
                f"Using AWS credentials for ARN: {identity['Arn']} "
                f"(UserId: {identity['UserId']}, Account: {identity['Account']})"
            )
        except Exception as e:
            logger.warning(f'Could not fetch caller identity: {e}')

        return True, None

    # IN GCS.PY
    @classmethod
    def get_k8s_credential_name(cls) -> str:
        return cls._AWS_SECRET_NAME


class S3CloudStorage(storage_utils.CloudStorage):
    """S3 Storage."""

    # List of commands to install AWS CLI
    _GET_AWSCLI = [
        'command -v aws >/dev/null 2>&1 || ('
        'apt-get update && apt-get install -y curl unzip && '
        'curl "https://awscli.amazonaws.com/awscli-exe-linux-x86_64.zip" -o "awscliv2.zip" && '  # noqa: E501
        'unzip awscliv2.zip && '
        './aws/install -i ~/aws-cli -b ~/bin && '
        'export PATH=$HOME/bin:$PATH && '
        'rm -rf aws awscliv2.zip'
        ') && export PATH=$HOME/bin:$PATH'
    ]

    _STORE: typing.Type[storage_utils.AbstractStore] = S3Store

    # IN GCS.PY
    def is_directory(self, url: str) -> bool:
        """Returns whether S3 'url' is a directory.

        In cloud object stores, a "directory" refers to a regular object whose
        name is a prefix of other objects.
        """
        s3 = aws.resource('s3')
        bucket_name, path = data_utils.split_s3_path(url)
        bucket = s3.Bucket(bucket_name)

        num_objects = 0
        for obj in bucket.objects.filter(Prefix=path):
            num_objects += 1
            if obj.key == path:
                return False
            # If there are more than 1 object in filter, then it is a directory
            if num_objects == 3:
                return True

        # A directory with few or no items
        return True

    # IN GCS.PY
    def make_sync_dir_command(self, source: str, destination: str) -> str:
        """Downloads using AWS CLI."""
        # AWS Sync by default uses 10 threads to upload files to the bucket.
        # To increase parallelism, modify max_concurrent_requests in your
        # aws config file (Default path: ~/.aws/config).
        all_commands = list(self._GET_AWSCLI)

        all_commands.append(f'aws s3 sync --no-follow-symlinks {source} {destination}')
        return ' && '.join(all_commands)

    # IN GCS.PY
    def make_sync_file_command(self, source: str, destination: str) -> str:
        all_commands = list(self._GET_AWSCLI)

        all_commands.append(f'aws s3 cp {source} {destination}')
        return ' && '.join(all_commands)



"""
Controller loop
The controller is run as a deployment and repeatedly polls the logging backend
for GPU related error logs. When a GPU, CUDA, NCCL error is detected,
we check against a set of known patterns(regexes) to see if the error is
irrecoverable making the node unfit for more work. We can then label/taint
the node to prevent more pods from being scheduled onto the node.

Sometimes an NCCL error can be raised due to all-reduce style workloads causing
all workers to fail even though only one is actually faulty. To place workers,
back into the working pool, we run a health check which just consists of doing NCCL
test on the tainted nodes
"""

import time
from typing import Set

from konduktor import logging
from konduktor.controller import constants, parse
from konduktor.controller import node as node_control

KONDUKTOR_CONTROLLER_LOG_POLL_SECONDS = 5
KONDUKTOR_CONTROLLER_HEALTH_CHECK_FREQ = 5

logger = logging.get_logger('konduktor.controller')


def main():
    logger.info(
        f'starting konduktor.controller ver. {constants.KONDUKTOR_CONTROLLER_VERSION}'
    )
    while True:
        for _ in range(KONDUKTOR_CONTROLLER_HEALTH_CHECK_FREQ):
            time.sleep(KONDUKTOR_CONTROLLER_LOG_POLL_SECONDS)
            error_by_pod: Set[str] = parse.pod_errors()
            error_by_dmesg: Set[str] = parse.dmesg_errors()
            for node in error_by_pod | error_by_dmesg:
                node_control.taint(node)

        node_control.health_check()


if __name__ == '__main__':
    main()



KONDUKTOR_IGNORE_FILE = '.konduktorignore'
GIT_IGNORE_FILE = '.gitignore'
KONDUKTOR_REMOTE_WORKDIR = '~/konduktor_workdir'
KONDUKTOR_LOGS_DIRECTORY = '~/konduktor_logs'

# Used for translate local file mounts to cloud storage. Please refer to
# konduktor/utils/controller_utils.py::maybe_translate_local_file_mounts_and_sync_up for
# more details.
# TODO(asaiacai): Unlike skypilot, we don't delete buckets after a job completes
# because we want to persists code, logs, and artifacts for debugging.
# yes it's a resource leak, but object store is
# so cheap and code/data is small in comparison.
FILE_MOUNTS_BUCKET_NAME = 'konduktor-filemounts-{username}-{user_hash}'
FILE_MOUNTS_LOCAL_TMP_DIR = 'konduktor-filemounts-files-{id}'
FILE_MOUNTS_REMOTE_TMP_DIR = '/tmp/konduktor-{}-filemounts-files'

# For API server, the use a temporary directory in the same path as the upload
# directory to avoid using a different block device, which may not allow hard
# linking. E.g., in our API server deployment on k8s, ~/.konduktor/ is mounted from a
# persistent volume, so any contents in ~/.konduktor/ cannot be hard linked elsewhere.
FILE_MOUNTS_LOCAL_TMP_BASE_PATH = '~/.konduktor/tmp/'
# Base path for two-hop file mounts translation. See
# controller_utils.translate_local_file_mounts_to_two_hop().
FILE_MOUNTS_CONTROLLER_TMP_BASE_PATH = '~/.konduktor/tmp/controller'


# Used when an managed jobs are created and
# files are synced up to the cloud.
FILE_MOUNTS_WORKDIR_SUBPATH = '{task_name}-{run_id}/workdir'
FILE_MOUNTS_SUBPATH = '{task_name}-{run_id}/local-file-mounts/{i}'
FILE_MOUNTS_TMP_SUBPATH = '{task_name}-{run_id}/tmp-files'

# Path to the file that contains the python path.
GET_PYTHON_PATH_CMD = 'which python3'
# Python executable, e.g., /opt/conda/bin/python3
PYTHON_CMD = f'$({GET_PYTHON_PATH_CMD})'



# Proprietary Changes made for Trainy under the Trainy Software License
# Original source: skypilot: https://github.com/skypilot-org/skypilot
# which is Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
On module import, we attempt to parse the config located at KONDUKTOR_CONFIG
(default: ~/.konduktor/config.yaml). Caller can then use

  >> konduktor_config.loaded()

to check if the config is successfully loaded.

To read a nested-key config:

  >> konduktor_config.get_nested(('auth', 'some_auth_config'), default_value)

The config can be overridden by the configs in task YAMLs. Callers are
responsible to provide the override_configs. If the nested key is part of
OVERRIDEABLE_CONFIG_KEYS, override_configs must be provided (can be empty):

  >> konduktor_config.get_nested(('docker', 'run_options'), default_value
                        override_configs={'docker': {'run_options': 'value'}})

To set a value in the nested-key config:

  >> config_dict = konduktor_config.set_nested(('auth', 'some_key'), value)

This operation returns a deep-copy dict, and is safe in that any key not found
will not raise an error.

Example usage:

Consider the following config contents:

    a:
        nested: 1
    b: 2

then:

    # Assuming ~/.konduktor/config.yaml exists and can be loaded:
    konduktor_config.loaded()  # ==> True

    konduktor_config.get_nested(('a', 'nested'), None)    # ==> 1
    konduktor_config.get_nested(('a', 'nonexist'), None)  # ==> None
    konduktor_config.get_nested(('a',), None)             # ==> {'nested': 1}

    # If ~/.konduktor/config.yaml doesn't exist or failed to be loaded:
    konduktor_config.loaded()  # ==> False
    konduktor_config.get_nested(('a', 'nested'), None)    # ==> None
    konduktor_config.get_nested(('a', 'nonexist'), None)  # ==> None
    konduktor_config.get_nested(('a',), None)             # ==> None
"""

import copy
import os
import pprint
from typing import Any, Dict, List, Optional, Tuple

import yaml

from konduktor import logging
from konduktor.utils import common_utils, schemas, ux_utils

logger = logging.get_logger(__name__)

# overrides are specified in task YAMLs.
OVERRIDEABLE_CONFIG_KEYS: List[Tuple[str, ...]] = [
    ('kubernetes', 'pod_config'),
]

# The config path is discovered in this order:
#
# (1) (Used internally) If env var {ENV_VAR_SKYPILOT_CONFIG} exists, use its
#     path;
# (2) If file {CONFIG_PATH} exists, use this file.
#
# If the path discovered by (1) fails to load, we do not attempt to go to step
# 2 in the list.

# (Used internally) An env var holding the path to the local config file. This
# is only used by jobs controller tasks to ensure recoveries of the same job
# use the same config file.
ENV_VAR_CONFIG = 'KONDUKTOR_CONFIG'

# Path to the local config file.
CONFIG_PATH = '~/.konduktor/config.yaml'


class Config(Dict[str, Any]):
    """Konduktor config that supports setting/getting values with nested keys."""

    def get_nested(
        self,
        keys: Tuple[str, ...],
        default_value: Any,
        override_configs: Optional[Dict[str, Any]] = None,
    ) -> Any:
        """Gets a nested key.

        If any key is not found, or any intermediate key does not point to a
        dict value, returns 'default_value'.

        Args:
            keys: A tuple of strings representing the nested keys.
            default_value: The default value to return if the key is not found.
            override_configs: A dict of override configs with the same schema as
                the config file, but only containing the keys to override.

        Returns:
            The value of the nested key, or 'default_value' if not found.
        """
        config = copy.deepcopy(self)
        if override_configs is not None:
            config = _recursive_update(config, override_configs)
        return _get_nested(config, keys, default_value)

    def set_nested(self, keys: Tuple[str, ...], value: Any) -> None:
        """In-place sets a nested key to value.

        Like get_nested(), if any key is not found, this will not raise an
        error.
        """
        override = {}
        for i, key in enumerate(reversed(keys)):
            if i == 0:
                override = {key: value}
            else:
                override = {key: override}
        _recursive_update(self, override)

    @classmethod
    def from_dict(cls, config: Optional[Dict[str, Any]]) -> 'Config':
        if config is None:
            return cls()
        return cls(**config)


# The loaded config.
_dict = Config()
_loaded_config_path: Optional[str] = None


def get_nested(
    keys: Tuple[str, ...],
    default_value: Any,
    override_configs: Optional[Dict[str, Any]] = None,
) -> Any:
    """Gets a nested key.

    If any key is not found, or any intermediate key does not point to a dict
    value, returns 'default_value'.

    When 'keys' is within OVERRIDEABLE_CONFIG_KEYS, 'override_configs' must be
    provided (can be empty). Otherwise, 'override_configs' must not be provided.

    Args:
        keys: A tuple of strings representing the nested keys.
        default_value: The default value to return if the key is not found.
        override_configs: A dict of override configs with the same schema as
            the config file, but only containing the keys to override.

    Returns:
        The value of the nested key, or 'default_value' if not found.
    """
    assert not (keys in OVERRIDEABLE_CONFIG_KEYS and override_configs is None), (
        f'Override configs must be provided when keys {keys} is within '
        'OVERRIDEABLE_CONFIG_KEYS: '
        f'{OVERRIDEABLE_CONFIG_KEYS}'
    )
    assert not (
        keys not in OVERRIDEABLE_CONFIG_KEYS and override_configs is not None
    ), (
        f'Override configs must not be provided when keys {keys} is not within '
        'OVERRIDEABLE_CONFIG_KEYS: '
        f'{OVERRIDEABLE_CONFIG_KEYS}'
    )
    return _dict.get_nested(keys, default_value, override_configs)


def set_nested(keys: Tuple[str, ...], value: Any) -> Dict[str, Any]:
    """Returns a deep-copied config with the nested key set to value.

    Like get_nested(), if any key is not found, this will not raise an error.
    """
    copied_dict = copy.deepcopy(_dict)
    copied_dict.set_nested(keys, value)
    return dict(**copied_dict)


def to_dict() -> Config:
    """Returns a deep-copied version of the current config."""
    return copy.deepcopy(_dict)


def _try_load_config() -> None:
    global _dict, _loaded_config_path
    config_path_via_env_var = os.environ.get(ENV_VAR_CONFIG)
    if config_path_via_env_var is not None:
        config_path = os.path.expanduser(config_path_via_env_var)
        if not os.path.exists(config_path):
            with ux_utils.print_exception_no_traceback():
                raise FileNotFoundError(
                    'Config file specified by env var '
                    f'{ENV_VAR_CONFIG} ({config_path!r}) does not '
                    'exist. Please double check the path or unset the env var: '
                    f'unset {ENV_VAR_CONFIG}'
                )
    else:
        config_path = CONFIG_PATH
    config_path = os.path.expanduser(config_path)
    if os.path.exists(config_path):
        logger.debug(f'Using config path: {config_path}')
        try:
            config = common_utils.read_yaml(config_path)
            _dict = Config.from_dict(config)
            _loaded_config_path = config_path
            logger.debug(f'Config loaded:\n{pprint.pformat(_dict)}')
        except yaml.YAMLError as e:
            logger.error(f'Error in loading config file ({config_path}):', e)
        if _dict:
            common_utils.validate_schema(
                _dict,
                schemas.get_config_schema(),
                f'Invalid config YAML ({config_path}). See: '
                'https://konduktor.readthedocs.io/en/latest/reference/config.html. '  # pylint: disable=line-too-long
                'Error: ',
                skip_none=False,
            )

        logger.debug('Config syntax check passed.')


def _check_allowed_and_disallowed_override_keys(
    key: str,
    allowed_override_keys: Optional[List[Tuple[str, ...]]] = None,
    disallowed_override_keys: Optional[List[Tuple[str, ...]]] = None,
) -> Tuple[Optional[List[Tuple[str, ...]]], Optional[List[Tuple[str, ...]]]]:
    allowed_keys_with_matched_prefix: Optional[List[Tuple[str, ...]]] = []
    disallowed_keys_with_matched_prefix: Optional[List[Tuple[str, ...]]] = []
    if allowed_override_keys is not None:
        for nested_key in allowed_override_keys:
            if key == nested_key[0]:
                if len(nested_key) == 1:
                    # Allowed key is fully matched, no need to check further.
                    allowed_keys_with_matched_prefix = None
                    break
                assert allowed_keys_with_matched_prefix is not None
                allowed_keys_with_matched_prefix.append(nested_key[1:])
        if (
            allowed_keys_with_matched_prefix is not None
            and not allowed_keys_with_matched_prefix
        ):
            raise ValueError(
                f'Key {key} is not in allowed override keys: '
                f'{allowed_override_keys}'
            )
    else:
        allowed_keys_with_matched_prefix = None

    if disallowed_override_keys is not None:
        for nested_key in disallowed_override_keys:
            if key == nested_key[0]:
                if len(nested_key) == 1:
                    raise ValueError(
                        f'Key {key} is in disallowed override keys: '
                        f'{disallowed_override_keys}'
                    )
                assert disallowed_keys_with_matched_prefix is not None
                disallowed_keys_with_matched_prefix.append(nested_key[1:])
    else:
        disallowed_keys_with_matched_prefix = None
    return allowed_keys_with_matched_prefix, disallowed_keys_with_matched_prefix


def _recursive_update(
    base_config: Config,
    override_config: Dict[str, Any],
    allowed_override_keys: Optional[List[Tuple[str, ...]]] = None,
    disallowed_override_keys: Optional[List[Tuple[str, ...]]] = None,
) -> Config:
    """Recursively updates base configuration with override configuration"""
    for key, value in override_config.items():
        (next_allowed_override_keys, next_disallowed_override_keys) = (
            _check_allowed_and_disallowed_override_keys(
                key, allowed_override_keys, disallowed_override_keys
            )
        )
        if key == 'kubernetes' and key in base_config:
            merge_k8s_configs(
                base_config[key],
                value,
                next_allowed_override_keys,
                next_disallowed_override_keys,
            )
        elif (
            isinstance(value, dict)
            and key in base_config
            and isinstance(base_config[key], dict)
        ):
            _recursive_update(
                base_config[key],
                value,
                next_allowed_override_keys,
                next_disallowed_override_keys,
            )
        else:
            base_config[key] = value
    return base_config


def _get_nested(
    configs: Optional[Dict[str, Any]],
    keys: Tuple[str, ...],
    default_value: Any,
    pop: bool = False,
) -> Any:
    if configs is None:
        return default_value
    curr = configs
    for i, key in enumerate(keys):
        if isinstance(curr, dict) and key in curr:
            value = curr[key]
            if i == len(keys) - 1:
                if pop:
                    curr.pop(key, default_value)
            curr = value
        else:
            return default_value
    logger.debug(f'User config: {".".join(keys)} -> {curr}')
    return curr


def merge_k8s_configs(
    base_config: Dict[Any, Any],
    override_config: Dict[Any, Any],
    allowed_override_keys: Optional[List[Tuple[str, ...]]] = None,
    disallowed_override_keys: Optional[List[Tuple[str, ...]]] = None,
) -> None:
    """Merge two configs into the base_config.

    Updates nested dictionaries instead of replacing them.
    If a list is encountered, it will be appended to the base_config list.

    An exception is when the key is 'containers', in which case the
    first container in the list will be fetched and merge_dict will be
    called on it with the first container in the base_config list.
    """
    for key, value in override_config.items():
        (next_allowed_override_keys, next_disallowed_override_keys) = (
            _check_allowed_and_disallowed_override_keys(
                key, allowed_override_keys, disallowed_override_keys
            )
        )
        if isinstance(value, dict) and key in base_config:
            merge_k8s_configs(
                base_config[key],
                value,
                next_allowed_override_keys,
                next_disallowed_override_keys,
            )
        elif isinstance(value, list) and key in base_config:
            assert isinstance(
                base_config[key], list
            ), f'Expected {key} to be a list, found {base_config[key]}'
            if key in ['containers', 'imagePullSecrets']:
                # If the key is 'containers' or 'imagePullSecrets, we take the
                # first and only container/secret in the list and merge it, as
                # we only support one container per pod.
                assert len(value) == 1, f'Expected only one container, found {value}'
                merge_k8s_configs(
                    base_config[key][0],
                    value[0],
                    next_allowed_override_keys,
                    next_disallowed_override_keys,
                )
            elif key in ['volumes', 'volumeMounts']:
                # If the key is 'volumes' or 'volumeMounts', we search for
                # item with the same name and merge it.
                for new_volume in value:
                    new_volume_name = new_volume.get('name')
                    if new_volume_name is not None:
                        destination_volume = next(
                            (
                                v
                                for v in base_config[key]
                                if v.get('name') == new_volume_name
                            ),
                            None,
                        )
                        if destination_volume is not None:
                            merge_k8s_configs(destination_volume, new_volume)
                        else:
                            base_config[key].append(new_volume)
            else:
                base_config[key].extend(value)
        else:
            base_config[key] = value


def loaded_config_path() -> Optional[str]:
    """Returns the path to the loaded config file."""
    return _loaded_config_path


# Load on import.
_try_load_config()


def loaded() -> bool:
    """Returns if the user configurations are loaded."""
    return bool(_dict)



# Proprietary Changes made for Trainy under the Trainy Software License
# Original source: skypilot: https://github.com/skypilot-org/skypilot
# which is Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Different cloud storage definitions. This modules responsibility
1.) Create the secrets for each cloud as k8s secrets
2.) Mount the secrets as volumes into each container
3.) Provide utilities/scripts for the pods to download files syncd
    to object storage

For each cloud/storage class we'll only have a single namespace at
`konduktor` and each run will correspond to a new folder e.g.
`s3://konduktor/my-llm-run-a34be-a3ebf`
"""

import enum
import os
import re
import urllib.parse
from typing import Any, Dict, List, Literal, Optional, Tuple, Type, Union

from konduktor import check, config, logging
from konduktor.data import aws, constants, data_utils, gcp, registry, storage_utils
from konduktor.utils import annotations, common_utils, exceptions, schemas, ux_utils

logger = logging.get_logger(__file__)


@annotations.lru_cache(scope='global')
def get_cached_enabled_storage_clouds_or_refresh(
    raise_if_no_cloud_access: bool = False,
) -> List[str]:
    # This is a temporary solution until https://github.com/skypilot-org/skypilot/issues/1943 # noqa: E501
    # (asaiacai): This function does not do any actual checking right now.
    # this is temporary. In the future, we can cache to disk.
    # For now, we just print a warning to the user saying what
    # clouds are enabled and if the task fails to run `konduktor check`
    # to update the credentials.
    enabled_clouds = config.get_nested(('allowed_clouds',), [])
    if len(enabled_clouds) == 0:
        enabled_clouds = registry._STORE_ENABLED_CLOUDS
    else:
        enabled_clouds = [str(cloud) for cloud in enabled_clouds]
    logger.warning(
        f'Enabled storage clouds: {enabled_clouds}. Defaulting to '
        f'{enabled_clouds[0]}. If sync fails, '
        're-run `konduktor check` to verify credentials.'
    )
    return enabled_clouds


def _is_storage_cloud_enabled(
    cloud_name: str, try_fix_with_sky_check: bool = True
) -> bool:
    enabled_storage_clouds = get_cached_enabled_storage_clouds_or_refresh()
    if cloud_name in enabled_storage_clouds:
        return True
    if try_fix_with_sky_check:
        # TODO(zhwu): Only check the specified cloud to speed up.
        check.check(quiet=True)
        return _is_storage_cloud_enabled(cloud_name, try_fix_with_sky_check=False)
    return False


class StorageMode(enum.Enum):
    COPY = 'COPY'
    MOUNT = 'MOUNT'


class StoreType(enum.Enum):
    """Enum for the different types of stores."""

    GCS = 'GCS'
    S3 = 'S3'

    @classmethod
    def from_cloud(cls, cloud: str) -> 'StoreType':
        # these need to match the cloud store classes in konduktor/cloud_stores.py
        if cloud.lower() == 'gs':
            return StoreType.GCS
        elif cloud.lower() == 's3':
            return StoreType.S3
        else:
            with ux_utils.print_exception_no_traceback():
                raise ValueError(f'Unknown cloud: {cloud}')

    @classmethod
    def from_store(cls, store: 'storage_utils.AbstractStore') -> 'StoreType':
        if store.__repr__() == 'GcsStore':
            return StoreType.GCS
        elif store.__repr__() == 'S3Store':
            return StoreType.S3
        else:
            with ux_utils.print_exception_no_traceback():
                raise ValueError(f'Unknown store type: {store}')

    def store_prefix(self) -> str:
        if self == StoreType.GCS:
            return 'gs://'
        elif self == StoreType.S3:
            return 's3://'
        else:
            with ux_utils.print_exception_no_traceback():
                raise ValueError(f'Unknown store type: {self}')

    @classmethod
    def get_fields_from_store_url(
        cls, store_url: str
    ) -> Tuple['StoreType', str, str, Optional[str], Optional[str]]:
        """Returns the store type, bucket name, and sub path from
        a store URL, and the storage account name and region if applicable.

        Args:
            store_url: str; The store URL.
        """
        # The full path from the user config of IBM COS contains the region,
        # and Azure Blob Storage contains the storage account name, we need to
        # pass these information to the store constructor.
        storage_account_name = None
        region = None
        for store_type in StoreType:
            if store_url.startswith(store_type.store_prefix()):
                if store_type == StoreType.GCS:
                    bucket_name, sub_path = data_utils.split_gcs_path(store_url)
                elif store_type == StoreType.S3:
                    bucket_name, sub_path = data_utils.split_s3_path(store_url)
                return store_type, bucket_name, sub_path, storage_account_name, region
        raise ValueError(f'Unknown store URL: {store_url}')

    @classmethod
    def get_endpoint_url(cls, store: 'storage_utils.AbstractStore', path: str) -> str:
        """Generates the endpoint URL for a given store and path.

        Args:
            store: Store object implementing AbstractStore.
            path: Path within the store.

        Returns:
            Endpoint URL of the bucket as a string.
        """
        store_type = cls.from_store(store)
        bucket_endpoint_url = f'{store_type.store_prefix()}{path}'
        return bucket_endpoint_url


# this should match the above StoreType enum
STORE_TYPES = Literal[StoreType.GCS, StoreType.S3]


class Storage(object):
    """Storage objects handle persistent and large volume storage in the sky.

    Storage represents an abstract data store containing large data files
    required by the task. Compared to file_mounts, storage is faster and
    can persist across runs, requiring fewer uploads from your local machine.

    A storage object can be used in either MOUNT mode or COPY mode. In MOUNT
    mode (the default), the backing store is directly "mounted" to the remote
    VM, and files are fetched when accessed by the task and files written to the
    mount path are also written to the remote store. In COPY mode, the files are
    pre-fetched and cached on the local disk and writes are not replicated on
    the remote store.

    Behind the scenes, storage automatically uploads all data in the source
    to a backing object store in a particular cloud (S3/GCS/Azure Blob).

      Typical Usage: (See examples/playground/storage_playground.py)
        storage = Storage(name='imagenet-bucket', source='~/Documents/imagenet')

        # Move data to S3
        storage.add_store('S3')

        # Move data to Google Cloud Storage
        storage.add_store('GCS')

        # Delete Storage for both S3 and GCS
        storage.delete()
    """

    class StorageMetadata(object):
        """A pickle-able tuple of:

        - (required) Storage name.
        - (required) Source
        - (optional) Storage mode.
        - (optional) Set of stores managed by sky added to the Storage object
        """

        def __init__(
            self,
            *,
            storage_name: Optional[str],
            source: Optional[constants.SourceType],
            mode: Optional[StorageMode] = None,
            sky_stores: Optional[
                Dict[StoreType, 'storage_utils.AbstractStore.StoreMetadata']
            ] = None,
        ):
            assert storage_name is not None or source is not None
            self.storage_name = storage_name
            self.source = source
            self.mode = mode

            # Only stores managed by sky are stored here in the
            # global_user_state
            self.sky_stores = {} if sky_stores is None else sky_stores

        def __repr__(self):
            return (
                f'StorageMetadata('
                f'\n\tstorage_name={self.storage_name},'
                f'\n\tsource={self.source},'
                f'\n\tmode={self.mode},'
                f'\n\t{self.sky_stores}'
            )

        def add_store(self, store: 'storage_utils.AbstractStore') -> None:
            storetype = StoreType.from_store(store)
            self.sky_stores[storetype] = store.get_metadata()

        def remove_store(self, store: 'storage_utils.AbstractStore') -> None:
            storetype = StoreType.from_store(store)
            if storetype in self.sky_stores:
                del self.sky_stores[storetype]

    def __init__(
        self,
        name: Optional[str] = None,
        source: Optional[constants.SourceType] = None,
        stores: Optional[List[STORE_TYPES]] = None,
        persistent: Optional[bool] = True,
        mode: StorageMode = StorageMode.COPY,
        sync_on_reconstruction: Optional[bool] = True,
        _is_sky_managed: Optional[bool] = False,
        _bucket_sub_path: Optional[str] = None,
    ) -> None:
        """Initializes a Storage object.

        Three fields are required: the name of the storage, the source
        path where the data is initially located, and the default mount
        path where the data will be mounted to on the cloud.

        Storage object validation depends on the name, source and mount mode.
        There are four combinations possible for name and source inputs:

        - name is None, source is None: Underspecified storage object.
        - name is not None, source is None: If MOUNT mode, provision an empty
            bucket with name <name>. If COPY mode, raise error since source is
            required.
        - name is None, source is not None: If source is local, raise error
            since name is required to create destination bucket. If source is
            a bucket URL, use the source bucket as the backing store (if
            permissions allow, else raise error).
        - name is not None, source is not None: If source is local, upload the
            contents of the source path to <name> bucket. Create new bucket if
            required. If source is bucket url - raise error. Name should not be
            specified if the source is a URL; name will be inferred from source.

        Args:
          name: str; Name of the storage object. Typically used as the
            bucket name in backing object stores.
          source: str, List[str]; File path where the data is initially stored.
            Can be a single local path, a list of local paths, or a cloud URI
            (s3://, gs://, etc.). Local paths do not need to be absolute.
          stores: Optional; Specify pre-initialized stores (S3Store, GcsStore).
          persistent: bool; Whether to persist across sky launches.
          mode: StorageMode; Specify how the storage object is manifested on
            the remote VM. Can be either MOUNT or COPY. Defaults to MOUNT.
          sync_on_reconstruction: bool; Whether to sync the data if the storage
            object is found in the global_user_state and reconstructed from
            there. This is set to false when the Storage object is created not
            for direct use, e.g. for 'sky storage delete', or the storage is
            being re-used, e.g., for `sky start` on a stopped cluster.
          _is_sky_managed: Optional[bool]; Indicates if the storage is managed
            by Sky. Without this argument, the controller's behavior differs
            from the local machine. For example, if a bucket does not exist:
            Local Machine (is_sky_managed=True) →
            Controller (is_sky_managed=False).
            With this argument, the controller aligns with the local machine,
            ensuring it retains the is_sky_managed information from the YAML.
            During teardown, if is_sky_managed is True, the controller should
            delete the bucket. Otherwise, it might mistakenly delete only the
            sub-path, assuming is_sky_managed is False.
          _bucket_sub_path: Optional[str]; The subdirectory to use for the
            storage object.
        """
        self.name: str
        self.source = source
        self.persistent = persistent
        self.mode = mode
        assert mode in StorageMode
        self.stores: Dict[StoreType, Optional['storage_utils.AbstractStore']] = {}
        if stores is not None:
            for store in stores:
                self.stores[store] = None
        self.sync_on_reconstruction = sync_on_reconstruction
        self._is_sky_managed = _is_sky_managed
        self._bucket_sub_path = _bucket_sub_path

        # TODO(romilb, zhwu): This is a workaround to support storage deletion
        # for spot. Once sky storage supports forced management for external
        # buckets, this can be deprecated.
        self.force_delete = False

        # Validate and correct inputs if necessary
        self._validate_storage_spec(name)

        if self.source is not None:
            # If source is a pre-existing bucket, connect to the bucket
            # If the bucket does not exist, this will error out
            if isinstance(self.source, str):
                if self.source.startswith('gs://'):
                    self.add_store(StoreType.GCS)
                elif self.source.startswith('s3://'):
                    self.add_store(StoreType.S3)

    @staticmethod
    def _validate_source(
        source: constants.SourceType,
        mode: StorageMode,
        sync_on_reconstruction: Optional[bool] = None,
    ) -> Tuple[constants.SourceType, bool]:
        """Validates the source path.

        Args:
          source: str; File path where the data is initially stored. Can be a
            local path or a cloud URI (s3://, gs://, r2:// etc.).
            Local paths do not need to be absolute.
          mode: StorageMode; StorageMode of the storage object

        Returns:
          Tuple[source, is_local_source]
          source: str; The source path.
          is_local_path: bool; Whether the source is a local path. False if URI.
        """

        def _check_basename_conflicts(source_list: List[str]) -> None:
            """Checks if two paths in source_list have the same basename."""
            basenames = [os.path.basename(s) for s in source_list]
            conflicts = {x for x in basenames if basenames.count(x) > 1}
            if conflicts:
                with ux_utils.print_exception_no_traceback():
                    raise exceptions.StorageSourceError(
                        'Cannot have multiple files or directories with the '
                        'same name in source. Conflicts found for: '
                        f'{", ".join(conflicts)}'
                    )

        def _validate_local_source(local_source):
            if local_source.endswith('/'):
                with ux_utils.print_exception_no_traceback():
                    raise exceptions.StorageSourceError(
                        'Storage source paths cannot end with a slash '
                        '(try "/mydir: /mydir" or "/myfile: /myfile"). '
                        f'Found source={local_source}'
                    )
            # Local path, check if it exists
            full_src = os.path.abspath(os.path.expanduser(local_source))
            # Only check if local source exists if it is synced to the bucket
            if not os.path.exists(full_src) and sync_on_reconstruction:
                with ux_utils.print_exception_no_traceback():
                    raise exceptions.StorageSourceError(
                        'Local source path does not' f' exist: {local_source}'
                    )
            # Raise warning if user's path is a symlink
            elif os.path.islink(full_src):
                logger.warning(
                    f'Source path {source} is a symlink. '
                    'Referenced contents are uploaded, matching '
                    'the default behavior for S3 and GCS syncing.'
                )

        # Check if source is a list of paths
        if isinstance(source, list):
            # Check for conflicts in basenames
            _check_basename_conflicts(source)
            # Validate each path
            for local_source in source:
                _validate_local_source(local_source)
            is_local_source = True
        else:
            # Check if str source is a valid local/remote URL
            split_path = urllib.parse.urlsplit(source)
            if split_path.scheme == '':
                _validate_local_source(source)
                # Check if source is a file - throw error if it is
                full_src = os.path.abspath(os.path.expanduser(source))
                if os.path.isfile(full_src):
                    with ux_utils.print_exception_no_traceback():
                        raise exceptions.StorageSourceError(
                            'Storage source path cannot be a file - only'
                            ' directories are supported as a source. '
                            'To upload a single file, specify it in a list '
                            f'by writing source: [{source}]. Note '
                            'that the file will be uploaded to the root of the '
                            'bucket and will appear at <destination_path>/'
                            f'{os.path.basename(source)}. Alternatively, you '
                            'can directly upload the file to the VM without '
                            'using a bucket by writing <destination_path>: '
                            f'{source} in the file_mounts section of your YAML'
                        )
                is_local_source = True
            elif split_path.scheme in ['s3', 'gs', 'https', 'r2', 'cos']:
                is_local_source = False
                # Storage mounting does not support mounting specific files from
                # cloud store - ensure path points to only a directory
                if mode == StorageMode.MOUNT:
                    if split_path.scheme != 'https' and (
                        (
                            split_path.scheme != 'cos'
                            and split_path.path.strip('/') != ''
                        )
                        or (
                            split_path.scheme == 'cos'
                            and not re.match(r'^/[-\w]+(/\s*)?$', split_path.path)
                        )
                    ):
                        # regex allows split_path.path to include /bucket
                        # or /bucket/optional_whitespaces while considering
                        # cos URI's regions (cos://region/bucket_name)
                        with ux_utils.print_exception_no_traceback():
                            raise exceptions.StorageModeError(
                                'MOUNT mode does not support'
                                ' mounting specific files from cloud'
                                ' storage. Please use COPY mode or'
                                ' specify only the bucket name as'
                                ' the source.'
                            )
            else:
                with ux_utils.print_exception_no_traceback():
                    raise exceptions.StorageSourceError(
                        f'Supported paths: local, s3://, gs://, https://, '
                        f'r2://, cos://. Got: {source}'
                    )
        return source, is_local_source

    def _validate_storage_spec(self, name: Optional[str]) -> None:
        """Validates the storage spec and updates local fields if necessary."""

        def validate_name(name):
            """Checks for validating the storage name.

            Checks if the name starts the s3, gcs or r2 prefix and raise error
            if it does. Store specific validation checks (e.g., S3 specific
            rules) happen in the corresponding store class.
            """
            prefix = name.split('://')[0]
            prefix = prefix.lower()
            if prefix in ['s3', 'gs', 'https', 'r2', 'cos']:
                with ux_utils.print_exception_no_traceback():
                    raise exceptions.StorageNameError(
                        'Prefix detected: `name` cannot start with '
                        f'{prefix}://. If you are trying to use an existing '
                        'bucket created outside of SkyPilot, please specify it '
                        'using the `source` field (e.g. '
                        '`source: s3://mybucket/`). If you are trying to '
                        'create a new bucket, please use the `store` field to '
                        'specify the store type (e.g. `store: s3`).'
                    )

        if self.source is None:
            # If the mode is COPY, the source must be specified
            if self.mode == StorageMode.COPY:
                # Check if a Storage object already exists in global_user_state
                # (e.g. used as scratch previously). Such storage objects can be
                # mounted in copy mode even though they have no source in the
                # yaml spec (the name is the source).
                # TODO(asaiacai): remove references to global_user_state
                # handle = global_user_state.get_handle_from_storage_name(name)
                handle = None
                if handle is None:
                    with ux_utils.print_exception_no_traceback():
                        raise exceptions.StorageSourceError(
                            'New storage object: source must be specified when '
                            'using COPY mode.'
                        )
            else:
                # If source is not specified in COPY mode, the intent is to
                # create a bucket and use it as scratch disk. Name must be
                # specified to create bucket.
                if not name:
                    with ux_utils.print_exception_no_traceback():
                        raise exceptions.StorageSpecError(
                            'Storage source or storage name must be specified.'
                        )
            assert name is not None, handle
            validate_name(name)
            self.name = name
            return
        elif self.source is not None:
            source, is_local_source = Storage._validate_source(
                self.source, self.mode, self.sync_on_reconstruction
            )
            if not name:
                if is_local_source:
                    with ux_utils.print_exception_no_traceback():
                        raise exceptions.StorageNameError(
                            'Storage name must be specified if the source is ' 'local.'
                        )
                else:
                    assert isinstance(source, str)
                    # Set name to source bucket name and continue
                    name = urllib.parse.urlsplit(source).netloc
                    assert name is not None, source
                    self.name = name
                    return
            else:
                if is_local_source:
                    # If name is specified and source is local, upload to bucket
                    assert name is not None, source
                    validate_name(name)
                    self.name = name
                    return
                else:
                    # Both name and source should not be specified if the source
                    # is a URI. Name will be inferred from the URI.
                    with ux_utils.print_exception_no_traceback():
                        raise exceptions.StorageSpecError(
                            'Storage name should not be specified if the '
                            'source is a remote URI.'
                        )
        raise exceptions.StorageSpecError(
            f'Validation failed for storage source {self.source}, name '
            f'{self.name} and mode {self.mode}. Please check the arguments.'
        )

    def _add_store_from_metadata(
        self, sky_stores: Dict[StoreType, 'storage_utils.AbstractStore.StoreMetadata']
    ) -> None:
        """Reconstructs Storage.stores from sky_stores.

        Reconstruct AbstractStore objects from sky_store's metadata and
        adds them into Storage.stores
        """
        for s_type, s_metadata in sky_stores.items():
            # When initializing from global_user_state, we override the
            # source from the YAML
            try:
                if s_type == StoreType.S3:
                    store = aws.S3Store.from_metadata(
                        s_metadata,
                        source=self.source,
                        sync_on_reconstruction=self.sync_on_reconstruction,
                    )
                elif s_type == StoreType.GCS:
                    store = gcp.GcsStore.from_metadata(
                        s_metadata,
                        source=self.source,
                        sync_on_reconstruction=self.sync_on_reconstruction,
                    )
                # elif s_type == StoreType.AZURE:
                #     assert isinstance(s_metadata,
                #                       AzureBlobStore.AzureBlobStoreMetadata)
                #     store = AzureBlobStore.from_metadata(
                #         s_metadata,
                #         source=self.source,
                #         sync_on_reconstruction=self.sync_on_reconstruction)
                # elif s_type == StoreType.R2:
                #     store = R2Store.from_metadata(
                #         s_metadata,
                #         source=self.source,
                #         sync_on_reconstruction=self.sync_on_reconstruction)
                # elif s_type == StoreType.IBM:
                #     store = IBMCosStore.from_metadata(
                #         s_metadata,
                #         source=self.source,
                #         sync_on_reconstruction=self.sync_on_reconstruction)
                else:
                    with ux_utils.print_exception_no_traceback():
                        raise ValueError(f'Unknown store type: {s_type}')
            # Following error is caught when an externally removed storage
            # is attempted to be fetched.
            except exceptions.StorageExternalDeletionError:
                logger.debug(
                    f'Storage object {self.name!r} was attempted '
                    'to be reconstructed while the corresponding '
                    'bucket was externally deleted.'
                )
                continue

            self._add_store(store, is_reconstructed=True)

    @classmethod
    def from_metadata(cls, metadata: StorageMetadata, **override_args) -> 'Storage':
        """Create Storage from StorageMetadata object.

        Used when reconstructing Storage object and AbstractStore objects from
        global_user_state.
        """
        # Name should not be specified if the source is a cloud store URL.
        source = override_args.get('source', metadata.source)
        name = override_args.get('name', metadata.storage_name)
        # If the source is a list, it consists of local paths
        if not isinstance(source, list) and data_utils.is_cloud_store_url(source):
            name = None

        storage_obj = cls(
            name=name,
            source=source,
            sync_on_reconstruction=override_args.get('sync_on_reconstruction', True),
        )

        # For backward compatibility
        if hasattr(metadata, 'mode'):
            if metadata.mode:
                storage_obj.mode = override_args.get('mode', metadata.mode)

        return storage_obj

    def add_store(
        self, store_type: Union[str, StoreType], region: Optional[str] = None
    ) -> 'storage_utils.AbstractStore':
        """Initializes and adds a new store to the storage.

        Invoked by the optimizer after it has selected a store to
        add it to Storage.

        Args:
          store_type: StoreType; Type of the storage [S3, GCS, AZURE, R2, IBM]
          region: str; Region to place the bucket in. Caller must ensure that
            the region is valid for the chosen store_type.
        """
        if isinstance(store_type, str):
            store_type = StoreType(store_type)

        store_cls: Type['storage_utils.AbstractStore']
        if store_type == StoreType.GCS:
            store_cls = gcp.GcsStore
        elif store_type == StoreType.S3:
            store_cls = aws.S3Store
        else:
            with ux_utils.print_exception_no_traceback():
                raise exceptions.StorageSpecError(
                    f'{store_type} not supported as a Store.'
                )

        # Initialize store object and get/create bucket
        try:
            assert self.source is not None
            store = store_cls(
                name=self.name,
                source=self.source,
                region=region,
                sync_on_reconstruction=self.sync_on_reconstruction,
                is_sky_managed=self._is_sky_managed,
                _bucket_sub_path=self._bucket_sub_path,
            )
        except exceptions.StorageBucketCreateError:
            # Creation failed, so this must be sky managed store. Add failure
            # to state.
            logger.error(
                f'Could not create {store_type} store ' f'with name {self.name}.'
            )
            raise
        except exceptions.StorageBucketGetError:
            # Bucket get failed, so this is not sky managed. Do not update state
            logger.error(f'Could not get {store_type} store ' f'with name {self.name}.')
            raise
        except exceptions.StorageInitError:
            logger.error(
                f'Could not initialize {store_type} store with '
                f'name {self.name}. General initialization error.'
            )
            raise
        except exceptions.StorageSpecError:
            logger.error(
                f'Could not mount externally created {store_type}'
                f'store with name {self.name!r}.'
            )
            raise

        # Add store to storage
        self._add_store(store)

        # Upload source to store
        self._sync_store(store)

        return store

    def _add_store(
        self, store: 'storage_utils.AbstractStore', is_reconstructed: bool = False
    ):
        # Adds a store object to the storage
        store_type = StoreType.from_store(store)
        self.stores[store_type] = store

    def delete(self, store_type: Optional[StoreType] = None) -> None:
        """Deletes data for all sky-managed storage objects.

        If a storage is not managed by sky, it is not deleted from the cloud.
        User must manually delete any object stores created outside of sky.

        Args:
            store_type: StoreType; Specific cloud store to remove from the list
              of backing stores.
        """
        if not self.stores:
            logger.info('No backing stores found. Deleting storage.')
        if store_type:
            store = self.stores[store_type]
            assert store is not None
            # We delete a store from the cloud if it's sky managed. Else just
            # remove handle and return
            if self.force_delete:
                store.delete()
            # Remove store from bookkeeping
            del self.stores[store_type]
        else:
            for _, store in self.stores.items():
                assert store is not None
                if self.force_delete:
                    store.delete()
            self.stores = {}

    def sync_all_stores(self):
        """Syncs the source and destinations of all stores in the Storage"""
        for _, store in self.stores.items():
            self._sync_store(store)

    def _sync_store(self, store: 'storage_utils.AbstractStore'):
        """Runs the upload routine for the store and handles failures"""

        def warn_for_git_dir(source: str):
            if os.path.isdir(os.path.join(source, '.git')):
                logger.warning(
                    f"'.git' directory under '{self.source}' "
                    'is excluded during sync.'
                )

        try:
            if self.source is not None:
                if isinstance(self.source, str):
                    warn_for_git_dir(self.source)
                else:
                    for source in self.source:
                        warn_for_git_dir(source)
            store.upload()
        except exceptions.StorageUploadError:
            logger.error(
                f'Could not upload {self.source!r} to store ' f'name {store.name!r}.'
            )
            raise

    @classmethod
    def from_yaml_config(cls, config: Dict[str, Any]) -> 'Storage':
        common_utils.validate_schema(
            config, schemas.get_storage_schema(), 'Invalid storage YAML: '
        )

        name = config.pop('name', None)
        source = config.pop('source', None)
        store = config.pop('store', None)
        mode_str = config.pop('mode', None)
        force_delete = config.pop('_force_delete', None)
        if force_delete is None:
            force_delete = False

        if isinstance(mode_str, str):
            # Make mode case insensitive, if specified
            mode = StorageMode(mode_str.upper())
        else:
            # Make sure this keeps the same as the default mode in __init__
            mode = StorageMode.MOUNT
        persistent = config.pop('persistent', None)
        if persistent is None:
            persistent = True

        assert not config, f'Invalid storage args: {config.keys()}'

        # Validation of the config object happens on instantiation.
        storage_obj = cls(name=name, source=source, persistent=persistent, mode=mode)
        if store is not None:
            storage_obj.add_store(StoreType(store.upper()))

        # Add force deletion flag
        storage_obj.force_delete = force_delete
        return storage_obj

    def to_yaml_config(self) -> Dict[str, str]:
        config = {}

        def add_if_not_none(key: str, value: Optional[Any]):
            if value is not None:
                config[key] = value

        name = None
        if (
            self.source is None
            or not isinstance(self.source, str)
            or not data_utils.is_cloud_store_url(self.source)
        ):
            # Remove name if source is a cloud store URL
            name = self.name
        add_if_not_none('name', name)
        add_if_not_none('source', self.source)

        stores = None
        if len(self.stores) > 0:
            stores = ','.join([store.value for store in self.stores])
        add_if_not_none('store', stores)
        add_if_not_none('persistent', self.persistent)
        add_if_not_none('mode', self.mode.value)
        if self.force_delete:
            config['_force_delete'] = True
        return config



# Proprietary Changes made for Trainy under the Trainy Software License
# Original source: skypilot: https://github.com/skypilot-org/skypilot
# which is Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Exceptions."""

import builtins
import traceback
import types
from typing import Any, Dict

# Return code for keyboard interruption and SIGTSTP
KEYBOARD_INTERRUPT_CODE = 130
SIGTSTP_CODE = 146
RSYNC_FILE_NOT_FOUND_CODE = 23
# Arbitrarily chosen value. Used in SkyPilot's storage mounting scripts
MOUNT_PATH_NON_EMPTY_CODE = 42
# Arbitrarily chosen value. Used to provision Kubernetes instance in Skypilot
INSUFFICIENT_PRIVILEGES_CODE = 52
# Return code when git command is ran in a dir that is not git repo
GIT_FATAL_EXIT_CODE = 128


def is_safe_exception(exc: Exception) -> bool:
    """Returns True if the exception is safe to send to clients.

    Safe exceptions are:
    1. Built-in exceptions
    2. Konduktor's own exceptions
    """
    module = type(exc).__module__

    # Builtin exceptions (e.g., ValueError, RuntimeError)
    if module == 'builtins':
        return True

    # Konduktor's own exceptions
    if module.startswith('sky.'):
        return True

    return False


def wrap_exception(exc: Exception) -> Exception:
    """Wraps non-safe exceptions into Konduktor exceptions

    This is used to wrap exceptions that are not safe to deserialize at clients.

    Examples include exceptions from cloud providers whose packages are not
    available at clients.
    """
    if is_safe_exception(exc):
        return exc

    return CloudError(
        message=str(exc),
        cloud_provider=type(exc).__module__.split('.')[0],
        error_type=type(exc).__name__,
    )


def serialize_exception(e: Exception) -> Dict[str, Any]:
    """Serialize the exception.

    This function also wraps any unsafe exceptions (e.g., cloud exceptions)
    into Konduktor's CloudError before serialization to ensure clients can
    deserialize them without needing cloud provider packages installed.
    """
    # Wrap unsafe exceptions before serialization
    e = wrap_exception(e)

    stacktrace = getattr(e, 'stacktrace', None)
    attributes = e.__dict__.copy()
    if 'stacktrace' in attributes:
        del attributes['stacktrace']
    for attr_k in list(attributes.keys()):
        attr_v = attributes[attr_k]
        if isinstance(attr_v, types.TracebackType):
            attributes[attr_k] = traceback.format_tb(attr_v)

    data = {
        'type': e.__class__.__name__,
        'message': str(e),
        'args': e.args,
        'attributes': attributes,
        'stacktrace': stacktrace,
    }
    return data


def deserialize_exception(serialized: Dict[str, Any]) -> Exception:
    """Deserialize the exception."""
    exception_type = serialized['type']
    if hasattr(builtins, exception_type):
        exception_class = getattr(builtins, exception_type)
    else:
        exception_class = globals().get(exception_type, None)
    if exception_class is None:
        # Unknown exception type.
        return Exception(f'{exception_type}: {serialized["message"]}')
    e = exception_class(*serialized['args'], **serialized['attributes'])
    if serialized['stacktrace'] is not None:
        setattr(e, 'stacktrace', serialized['stacktrace'])
    return e


class CloudError(Exception):
    """Wraps cloud-specific errors into a SkyPilot exception."""

    def __init__(self, message: str, cloud_provider: str, error_type: str):
        super().__init__(message)
        self.cloud_provider = cloud_provider
        self.error_type = error_type

    def __str__(self):
        return (
            f'{self.cloud_provider} error ({self.error_type}): ' f'{super().__str__()}'
        )


class CommandError(Exception):
    pass


class NotSupportedError(Exception):
    """Raised when a feature is not supported."""

    pass


class StorageError(Exception):
    pass


class StorageSpecError(ValueError):
    # Errors raised due to invalid specification of the Storage object
    pass


class StorageInitError(StorageError):
    # Error raised when Initialization fails - either due to permissions,
    # unavailable name, or other reasons.
    pass


class StorageBucketCreateError(StorageInitError):
    # Error raised when bucket creation fails.
    pass


class StorageBucketGetError(StorageInitError):
    # Error raised if attempt to fetch an existing bucket fails.
    pass


class StorageBucketDeleteError(StorageError):
    # Error raised if attempt to delete an existing bucket fails.
    pass


class StorageUploadError(StorageError):
    # Error raised when bucket is successfully initialized, but upload fails,
    # either due to permissions, ctrl-c, or other reasons.
    pass


class StorageSourceError(StorageSpecError):
    # Error raised when the source of the storage is invalid. E.g., does not
    # exist, malformed path, or other reasons.
    pass


class StorageNameError(StorageSpecError):
    # Error raised when the source of the storage is invalid. E.g., does not
    # exist, malformed path, or other reasons.
    pass


class StorageModeError(StorageSpecError):
    # Error raised when the storage mode is invalid or does not support the
    # requested operation (e.g., passing a file as source to MOUNT mode)
    pass


class StorageExternalDeletionError(StorageBucketGetError):
    # Error raised when the bucket is attempted to be fetched while it has been
    # deleted externally.
    pass


class NonExistentStorageAccountError(StorageExternalDeletionError):
    # Error raise when storage account provided through config.yaml or read
    # from store handle(local db) does not exist.
    pass


class NetworkError(Exception):
    """Raised when network fails."""

    pass


class CloudUserIdentityError(Exception):
    """Raised when the cloud identity is invalid."""

    pass


class ClusterOwnerIdentityMismatchError(Exception):
    """The cluster's owner identity does not match the current user identity."""

    pass


class NoCloudAccessError(Exception):
    """Raised when all clouds are disabled."""

    pass



"""Constants for usage collection."""

import os

KONDUKTOR_DISABLE_USAGE_COLLECTION = os.environ.get(
    'KONDUKTOR_DISABLE_USAGE_COLLECTION', False
)

POSTHOG_API_KEY = os.environ.get(
    'POSTHOG_API_KEY', 'phc_4UgX80BfVNmYRZ2o3dJLyRMGkv1CxBozPAcPnD29uP4'
)

POSTHOG_HOST = os.environ.get('POSTHOG_HOST', 'https://us.i.posthog.com')

USAGE_POLICY_MESSAGE = (
    'Konduktor collects usage data to improve its services. '
    '`run` commands are not collected to '
    'ensure privacy.\n'
    'Usage logging can be disabled by setting the '
    'environment variable KONDUKTOR_DISABLE_USAGE_COLLECTION=1.'
)



"""Data sync between workstation <--> blob (s3, gcs, etc.) <--> worker pods"""

from konduktor.data.aws.s3 import (
    DEFAULT_AWS_CONFIG_PATH,
    DEFAULT_AWS_CREDENTIAL_PATH,
    S3CloudStorage,
    S3Store,
)

__all__ = [
    'S3Store',
    'S3CloudStorage',
    'DEFAULT_AWS_CREDENTIAL_PATH',
    'DEFAULT_AWS_CONFIG_PATH',
]



# Proprietary Changes made for Trainy under the Trainy Software License
# Original source: konduktor: https://github.com/konduktor-org/konduktor
# which is Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Task: a coarse-grained stage in an application."""

import inspect
import json
import os
import re
import typing
from typing import Any, Dict, List, Optional, Tuple, Union

import yaml

if typing.TYPE_CHECKING:
    import konduktor.resource as resources_lib

import konduktor
from konduktor import constants, logging
from konduktor.data import data_utils
from konduktor.data import storage as storage_lib
from konduktor.utils import common_utils, exceptions, schemas, ux_utils

logger = logging.get_logger(__name__)

_VALID_NAME_REGEX = '[a-zA-Z0-9]+(?:[._-]{1,2}[a-zA-Z0-9]+)*'
_VALID_NAME_DESCR = (
    'ASCII characters and may contain lowercase and'
    ' uppercase letters, digits, underscores, periods,'
    ' and dashes. Must start and end with alphanumeric'
    ' characters. No triple dashes or underscores.'
)

_RUN_FN_CHECK_FAIL_MSG = (
    'run command generator must take exactly 2 arguments: node_rank (int) and'
    'a list of node ip addresses (List[str]). Got {run_sig}'
)


def _is_valid_name(name: Optional[str]) -> bool:
    """Checks if the task name is valid.

    Valid is defined as either NoneType or str with ASCII characters which may
    contain lowercase and uppercase letters, digits, underscores, periods,
    and dashes. Must start and end with alphanumeric characters.
    No triple dashes or underscores.

    Examples:
        some_name_here
        some-name-here
        some__name__here
        some--name--here
        some__name--here
        some.name.here
        some-name_he.re
        this---shouldnt--work
        this___shouldnt_work
        _thisshouldntwork
        thisshouldntwork_
    """
    if name is None:
        return False
    return bool(re.fullmatch(_VALID_NAME_REGEX, name))


def _fill_in_env_vars(
    yaml_field: Dict[str, Any],
    task_envs: Dict[str, str],
) -> Dict[str, Any]:
    """Detects env vars in yaml field and fills them with task_envs.

    Use cases of env vars in file_mounts:
    - dst/src paths; e.g.,
        /model_path/llama-${SIZE}b: s3://llama-weights/llama-${SIZE}b
    - storage's name (bucket name)
    - storage's source (local path)

    Use cases of env vars in service:
    - model type; e.g.,
        service:
          readiness_probe:
            path: /v1/chat/completions
            post_data:
              model: $MODEL_NAME
              messages:
                - role: user
                  content: How to print hello world?
              max_tokens: 1

    We simply dump yaml_field into a json string, and replace env vars using
    regex. This should be safe as yaml config has been schema-validated.

    Env vars of the following forms are detected:
        - ${ENV}
        - $ENV
    where <ENV> must appear in task.envs.
    """
    yaml_field_str = json.dumps(yaml_field)

    def replace_var(match):
        var_name = match.group(1)
        # If the variable isn't in the dictionary, return it unchanged
        return task_envs.get(var_name, match.group(0))

    # Pattern for valid env var names in bash.
    pattern = r'\$\{?\b([a-zA-Z_][a-zA-Z0-9_]*)\b\}?'
    yaml_field_str = re.sub(pattern, replace_var, yaml_field_str)
    return json.loads(yaml_field_str)


class Task:
    """Task: a computation to be run on the cloud."""

    def __init__(
        self,
        name: str,
        *,
        setup: Optional[str] = None,
        run: Optional[str] = None,
        envs: Optional[Dict[str, str]] = None,
        workdir: Optional[str] = None,
        num_nodes: Optional[int] = None,
    ):
        """Initializes a Task.

        All fields are optional.  ``Task.run`` is the actual program: either a
        shell command to run (str) or a command generator for different nodes
        (lambda; see below).

        Optionally, call ``Task.set_resources()`` to set the resource
        requirements for this task.  If not set, a default CPU-only requirement
        is assumed (the same as ``konduktor launch``).

        All setters of this class, ``Task.set_*()``, return ``self``, i.e.,
        they are fluent APIs and can be chained together.

        Example:
            .. code-block:: python

                # A Task that will sync up local workdir '.', containing
                # requirements.txt and train.py.
                konduktor.Task(setup='pip install requirements.txt',
                         run='python train.py',
                         workdir='.')

                # An empty Task for provisioning a cluster.
                task = konduktor.Task(num_nodes=n).set_resources(...)

                # Chaining setters.
                konduktor.Task().set_resources(...).set_file_mounts(...)

        Args:
          name: A string name for the Task for display purposes.
          run: The actual command for the task. If not None, either a shell
            command (str) or a command generator (callable).  If latter, it
            must take a node rank and a list of node addresses as input and
            return a shell command (str) (valid to return None for some nodes,
            in which case no commands are run on them).  Run commands will be
            run under ``workdir``. Note the command generator should be a
            self-contained lambda.
          envs: A dictionary of environment variables to set before running the
            setup and run commands.
          workdir: The local working directory.  This directory will be synced
            to a location on the remote VM(s), and ``setup`` and ``run``
            commands will be run under that location (thus, they can rely on
            relative paths when invoking binaries).
          num_nodes: The number of nodes to provision for this Task.  If None,
            treated as 1 node.  If > 1, each node will execute its own
            setup/run command, where ``run`` can either be a str, meaning all
            nodes get the same command, or a lambda, with the semantics
            documented above.
        """
        assert name is not None, 'Task name is required'
        self.name = name
        if setup is not None:
            raise ValueError('`setup` is being deprecated and not supported')
        self.run = run
        self.storage_mounts: Dict[str, storage_lib.Storage] = {}
        self.storage_plans: Dict[storage_lib.Storage, storage_lib.StoreType] = {}
        self._envs = envs or {}
        self.workdir = workdir
        # Ignore type error due to a mypy bug.
        # https://github.com/python/mypy/issues/3004
        self._num_nodes = 1
        self.num_nodes = num_nodes  # type: ignore
        # Default to CPU VM
        self.resources: Optional[konduktor.Resources] = None

        self.file_mounts: Optional[Dict[str, str]] = None
        self.best_resources = None  # (asaiacai): this is unused consider removing

        # Check if the task is legal.
        self._validate()

    def _validate(self):
        """Checks if the Task fields are valid."""

        # TODO(asaiacai): add validations here to check that valid
        # kueue resources are specified as labels
        if not _is_valid_name(self.name):
            with ux_utils.print_exception_no_traceback():
                raise ValueError(
                    f'Invalid task name {self.name}. Valid name: '
                    f'{_VALID_NAME_DESCR}'
                )

        # Check self.run
        if callable(self.run):
            run_sig = inspect.signature(self.run)
            # Check that run is a function with 2 arguments.
            if len(run_sig.parameters) != 2:
                with ux_utils.print_exception_no_traceback():
                    raise ValueError(_RUN_FN_CHECK_FAIL_MSG.format(run_sig))

            type_list = [int, List[str]]
            # Check annotations, if exists
            for i, param in enumerate(run_sig.parameters.values()):
                if param.annotation != inspect.Parameter.empty:
                    if param.annotation != type_list[i]:
                        with ux_utils.print_exception_no_traceback():
                            raise ValueError(_RUN_FN_CHECK_FAIL_MSG.format(run_sig))

            # Check self containedness.
            run_closure = inspect.getclosurevars(self.run)
            if run_closure.nonlocals:
                with ux_utils.print_exception_no_traceback():
                    raise ValueError(
                        'run command generator must be self contained. '
                        f'Found nonlocals: {run_closure.nonlocals}'
                    )
            if run_closure.globals:
                with ux_utils.print_exception_no_traceback():
                    raise ValueError(
                        'run command generator must be self contained. '
                        f'Found globals: {run_closure.globals}'
                    )
            if run_closure.unbound:
                # Do not raise an error here. Import statements, which are
                # allowed, will be considered as unbounded.
                pass
        elif self.run is not None and not isinstance(self.run, str):
            with ux_utils.print_exception_no_traceback():
                raise ValueError(
                    'run must be a shell script (str), ' f'Got {type(self.run)}'
                )
        elif self.run is None:
            with ux_utils.print_exception_no_traceback():
                raise ValueError('run commands are empty')

        # Workdir.
        if self.workdir is not None:
            full_workdir = os.path.abspath(os.path.expanduser(self.workdir))
            if not os.path.isdir(full_workdir):
                # Symlink to a dir is legal (isdir() follows symlinks).
                with ux_utils.print_exception_no_traceback():
                    raise ValueError(
                        'Workdir must exist and must be a directory (or '
                        f'a symlink to a directory). {self.workdir} not found.'
                    )

    @staticmethod
    def from_yaml_config(
        config: Dict[str, Any],
        env_overrides: Optional[List[Tuple[str, str]]] = None,
    ) -> 'Task':
        # More robust handling for 'envs': explicitly convert keys and values to
        # str, since users may pass '123' as keys/values which will get parsed
        # as int causing validate_schema() to fail.
        envs = config.get('envs')
        if envs is not None and isinstance(envs, dict):
            new_envs: Dict[str, Optional[str]] = {}
            for k, v in envs.items():
                if v is not None:
                    new_envs[str(k)] = str(v)
                else:
                    new_envs[str(k)] = None
            config['envs'] = new_envs
        common_utils.validate_schema(
            config, schemas.get_task_schema(), 'Invalid task YAML: '
        )
        if env_overrides is not None:
            # We must override env vars before constructing the Task, because
            # the Storage object creation is eager and it (its name/source
            # fields) may depend on env vars.
            new_envs = config.get('envs', {})
            new_envs.update(env_overrides)
            config['envs'] = new_envs

        for k, v in config.get('envs', {}).items():
            if v is None:
                with ux_utils.print_exception_no_traceback():
                    raise ValueError(
                        f'Environment variable {k!r} is None. Please set a '
                        'value for it in task YAML or with --env flag. '
                        f'To set it to be empty, use an empty string ({k}: "" '
                        f'in task YAML or --env {k}="" in CLI).'
                    )

        # Fill in any Task.envs into file_mounts (src/dst paths, storage
        # name/source).
        if config.get('file_mounts') is not None:
            config['file_mounts'] = _fill_in_env_vars(
                config['file_mounts'], config.get('envs', {})
            )

        # Fill in any Task.envs into workdir
        if config.get('workdir') is not None:
            config['workdir'] = _fill_in_env_vars(
                config['workdir'], config.get('envs', {})
            )

        task = Task(
            config.pop('name', None),
            run=config.pop('run', None),
            workdir=config.pop('workdir', None),
            num_nodes=config.pop('num_nodes', None),
            envs=config.pop('envs', None),
        )

        # Create lists to store storage objects inlined in file_mounts.
        # These are retained in dicts in the YAML schema and later parsed to
        # storage objects with the storage/storage_mount objects.
        fm_storages = []
        file_mounts = config.pop('file_mounts', None)
        if file_mounts is not None:
            copy_mounts = {}
            for dst_path, src in file_mounts.items():
                # Check if it is str path
                if isinstance(src, str):
                    copy_mounts[dst_path] = src
                # If the src is not a str path, it is likely a dict. Try to
                # parse storage object.
                elif isinstance(src, dict):
                    fm_storages.append((dst_path, src))
                else:
                    with ux_utils.print_exception_no_traceback():
                        raise ValueError(
                            f'Unable to parse file_mount ' f'{dst_path}:{src}'
                        )
            task.set_file_mounts(copy_mounts)

        task_storage_mounts: Dict[str, storage_lib.Storage] = {}
        all_storages = fm_storages
        for storage in all_storages:
            mount_path = storage[0]
            assert mount_path, 'Storage mount path cannot be empty.'
            try:
                storage_obj = storage_lib.Storage.from_yaml_config(storage[1])
            except exceptions.StorageSourceError as e:
                # Patch the error message to include the mount path, if included
                e.args = (
                    e.args[0].replace('<destination_path>', mount_path),
                ) + e.args[1:]
                raise e
            task_storage_mounts[mount_path] = storage_obj
        task.set_storage_mounts(task_storage_mounts)

        # Experimental configs.
        experimnetal_configs = config.pop('experimental', None)
        cluster_config_override = None
        if experimnetal_configs is not None:
            cluster_config_override = experimnetal_configs.pop('config_overrides', None)
            logger.debug(
                'Overriding konduktor config with task-level config: '
                f'{cluster_config_override}'
            )
        assert not experimnetal_configs, (
            'Invalid task args: ' f'{experimnetal_configs.keys()}'
        )

        # Parse resources field.
        resources_config = config.pop('resources', {})
        if cluster_config_override is not None:
            assert resources_config.get('_cluster_config_overrides') is None, (
                'Cannot set _cluster_config_overrides in both resources and '
                'experimental.config_overrides'
            )
            resources_config['_cluster_config_overrides'] = cluster_config_override
        task.set_resources(konduktor.Resources.from_yaml_config(resources_config))
        assert not config, f'Invalid task args: {config.keys()}'
        return task

    @staticmethod
    def from_yaml(yaml_path: str) -> 'Task':
        """Initializes a task from a task YAML.

        Example:
            .. code-block:: python

                task = konduktor.Task.from_yaml('/path/to/task.yaml')

        Args:
          yaml_path: file path to a valid task yaml file.

        Raises:
          ValueError: if the path gets loaded into a str instead of a dict; or
            if there are any other parsing errors.
        """
        with open(os.path.expanduser(yaml_path), 'r', encoding='utf-8') as f:
            #  https://github.com/yaml/pyyaml/issues/165#issuecomment-430074049
            # to raise errors on duplicate keys.
            config = yaml.safe_load(f)

        if isinstance(config, str):
            with ux_utils.print_exception_no_traceback():
                raise ValueError(
                    'YAML loaded as str, not as dict. '
                    f'Is it correct? Path: {yaml_path}'
                )

        if config is None:
            config = {}
        return Task.from_yaml_config(config)

    @property
    def num_nodes(self) -> int:
        return self._num_nodes

    @num_nodes.setter
    def num_nodes(self, num_nodes: Optional[int]) -> None:
        if num_nodes is None:
            num_nodes = 1
        if not isinstance(num_nodes, int) or num_nodes <= 0:
            with ux_utils.print_exception_no_traceback():
                raise ValueError(
                    f'num_nodes should be a positive int. Got: {num_nodes}'
                )
        self._num_nodes = num_nodes

    @property
    def envs(self) -> Dict[str, str]:
        return self._envs

    def update_envs(
        self, envs: Union[None, List[Tuple[str, str]], Dict[str, str]]
    ) -> 'Task':
        """Updates environment variables for use inside the setup/run commands.

        Args:
          envs: (optional) either a list of ``(env_name, value)`` or a dict
            ``{env_name: value}``.

        Returns:
          self: The current task, with envs updated.

        Raises:
          ValueError: if various invalid inputs errors are detected.
        """
        if envs is None:
            envs = {}
        if isinstance(envs, (list, tuple)):
            keys = set(env[0] for env in envs)
            if len(keys) != len(envs):
                with ux_utils.print_exception_no_traceback():
                    raise ValueError('Duplicate env keys provided.')
            envs = dict(envs)
        if isinstance(envs, dict):
            for key in envs:
                if not isinstance(key, str):
                    with ux_utils.print_exception_no_traceback():
                        raise ValueError('Env keys must be strings.')
                if not common_utils.is_valid_env_var(key):
                    with ux_utils.print_exception_no_traceback():
                        raise ValueError(f'Invalid env key: {key}')
        else:
            with ux_utils.print_exception_no_traceback():
                raise ValueError(
                    'envs must be List[Tuple[str, str]] or Dict[str, str]: ' f'{envs}'
                )
        self._envs.update(envs)
        return self

    def set_resources(
        self,
        resources: 'resources_lib.Resources',
    ) -> 'Task':
        """Sets the required resources to execute this task.

        If this function is not called for a Task, default resource
        requirements will be used (8 vCPUs).

        Args:
          resources: either a konduktor.Resources, a set of them, or a list of them.
            A set or a list of resources asks the optimizer to "pick the
            best of these resources" to run this task.
        Returns:
          self: The current task, with resources set.
        """
        if isinstance(resources, konduktor.Resources):
            resources = resources
        self.resources = resources

        # TODO(asaiacai): we're only going to support COPY for now.
        # MOUNT is not supported. Evaluate if the task requires
        # FUSE and set the requires_fuse flag
        for _, storage_obj in self.storage_mounts.items():
            if storage_obj.mode not in storage_lib.StorageMode:
                with ux_utils.print_exception_no_traceback():
                    raise ValueError(
                        f'Storage Type {storage_obj.mode} '
                        'not supported. '
                        f'Only {", ".join([mode.name for mode in storage_lib.StorageMode])} '  # noqa: E501
                        'is supported!'
                    )
            # if storage_obj.mode == storage_lib.StorageMode.MOUNT:
            #     for r in self.resources:
            #         r.requires_fuse = True
            #     break

        return self

    def set_resources_override(self, override_params: Dict[str, Any]) -> 'Task':
        """Sets the override parameters for the resources."""
        assert self.resources is not None, 'Resources are required'
        new_resources = self.resources.copy(**override_params)

        self.set_resources(new_resources)
        return self

    def set_file_mounts(self, file_mounts: Optional[Dict[str, str]]) -> 'Task':
        """Sets the file mounts for this task.

        Useful for syncing datasets, dotfiles, etc.

        File mounts are a dictionary: ``{remote_path: local_path/cloud URI}``.
        Local (or cloud) files/directories will be synced to the specified
        paths on the remote VM(s) where this Task will run.

        Neither source or destimation paths can end with a slash.

        Example:
            .. code-block:: python

                task.set_file_mounts({
                    '~/.dotfile': '/local/.dotfile',
                    # /remote/dir/ will contain the contents of /local/dir/.
                    '/remote/dir': '/local/dir',
                })

        Args:
          file_mounts: an optional dict of ``{remote_path: local_path/cloud
            URI}``, where remote means the VM(s) on which this Task will
            eventually run on, and local means the node from which the task is
            launched.

        Returns:
          self: the current task, with file mounts set.

        Raises:
          ValueError: if input paths are invalid.
        """
        if file_mounts is None:
            self.file_mounts = None
            return self
        for target, source in file_mounts.items():
            if target.endswith('/') or source.endswith('/'):
                with ux_utils.print_exception_no_traceback():
                    raise ValueError(
                        'File mount paths cannot end with a slash '
                        '(try "/mydir: /mydir" or "/myfile: /myfile"). '
                        f'Found: target={target} source={source}'
                    )
            if data_utils.is_cloud_store_url(target):
                with ux_utils.print_exception_no_traceback():
                    raise ValueError(
                        'File mount destination paths cannot be cloud storage'
                    )
            if not data_utils.is_cloud_store_url(source):
                if not os.path.exists(
                    os.path.abspath(os.path.expanduser(source))
                ) and not source.startswith('konduktor:'):
                    with ux_utils.print_exception_no_traceback():
                        raise ValueError(
                            f'File mount source {source!r} does not exist '
                            'locally. To fix: check if it exists, and correct '
                            'the path.'
                        )
            if (
                target == constants.KONDUKTOR_REMOTE_WORKDIR
                and self.workdir is not None
            ):
                with ux_utils.print_exception_no_traceback():
                    raise ValueError(
                        f'Cannot use {constants.KONDUKTOR_REMOTE_WORKDIR!r} as a '
                        'destination path of a file mount, as it will be used '
                        'by the workdir. If uploading a file/folder to the '
                        'workdir is needed, please specify the full path to '
                        'the file/folder.'
                    )

        self.file_mounts = file_mounts
        return self

    def _get_preferred_store(self) -> Tuple[storage_lib.StoreType, Optional[str]]:
        """Returns the preferred store type and region for this task."""
        storage_cloud = None

        enabled_storage_clouds = (
            storage_lib.get_cached_enabled_storage_clouds_or_refresh(
                raise_if_no_cloud_access=True
            )
        )
        resources = self.resources
        if resources is not None:
            storage_cloud = resources.cloud
        else:
            storage_cloud = None

        if storage_cloud is not None:
            if str(storage_cloud) not in enabled_storage_clouds:
                storage_cloud = None

        storage_cloud_str = None
        if storage_cloud is None:
            storage_cloud_str = enabled_storage_clouds[0]
            assert storage_cloud_str is not None, enabled_storage_clouds[0]
            storage_region = None  # Use default region in the Store class
        else:
            storage_cloud_str = str(storage_cloud)

        store_type = storage_lib.StoreType.from_cloud(storage_cloud_str)
        return store_type, storage_region

    def sync_storage_mounts(self) -> None:
        """(INTERNAL) Eagerly syncs storage mounts to cloud storage.

        After syncing up, COPY-mode storage mounts are translated into regular
        file_mounts of the form ``{ /remote/path: {s3,gs,..}://<bucket path>
        }``. For local file mounts, we first sync all local paths from
        `workdir` and `file_mounts` to the cloud storage.
        """
        for storage in self.storage_mounts.values():
            if len(storage.stores) == 0:
                store_type, store_region = self._get_preferred_store()
                self.storage_plans[storage] = store_type
                storage.add_store(store_type, store_region)
            else:
                # We will download the first store that is added to remote.
                self.storage_plans[storage] = list(storage.stores.keys())[0]

        storage_mounts = self.storage_mounts
        storage_plans = self.storage_plans
        for mnt_path, storage in storage_mounts.items():
            if storage.mode == storage_lib.StorageMode.COPY:
                store_type = storage_plans[storage]
            # TODO(asaiacai): add back other stores here
            elif store_type is storage_lib.StoreType.S3:
                if isinstance(storage.source, str) and storage.source.startswith(
                    's3://'
                ):
                    blob_path = storage.source
                else:
                    assert storage.name is not None, storage
                    blob_path = 's3://' + storage.name
                self.update_file_mounts(
                    {
                        mnt_path: blob_path,
                    }
                )
            elif store_type is storage_lib.StoreType.GCS:
                if isinstance(storage.source, str) and storage.source.startswith(
                    'gs://'
                ):
                    blob_path = storage.source
                else:
                    assert storage.name is not None, storage
                    blob_path = 'gs://' + storage.name
                self.update_file_mounts(
                    {
                        mnt_path: blob_path,
                    }
                )
            else:
                with ux_utils.print_exception_no_traceback():
                    raise ValueError(f'Storage Type {store_type} ' 'does not exist!')

    def update_storage_mounts(
        self, storage_mounts: Dict[str, storage_lib.Storage]
    ) -> 'Task':
        """Updates the storage mounts for this task.

        Different from set_storage_mounts(), this function updates into the
        existing storage_mounts (calls ``dict.update()``), rather than
        overwriting it.

        This should be called before provisioning in order to take effect.

        Args:
          storage_mounts: an optional dict of ``{mount_path: sky.Storage
            object}``, where mount_path is the path inside the remote VM(s)
            where the Storage object will be mounted on.

        Returns:
          self: The current task, with storage mounts updated.

        Raises:
          ValueError: if input paths are invalid.
        """
        if not storage_mounts:
            return self
        task_storage_mounts = self.storage_mounts if self.storage_mounts else {}
        task_storage_mounts.update(storage_mounts)
        return self.set_storage_mounts(task_storage_mounts)

    def update_file_mounts(self, file_mounts: Dict[str, str]) -> 'Task':
        """Updates the file mounts for this task.

        Different from set_file_mounts(), this function updates into the
        existing file_mounts (calls ``dict.update()``), rather than
        overwritting it.

        This should be called before provisioning in order to take effect.

        Example:
            .. code-block:: python

                task.update_file_mounts({
                    '~/.config': '~/Documents/config',
                    '/tmp/workdir': '/local/workdir/cnn-cifar10',
                })

        Args:
          file_mounts: a dict of ``{remote_path: local_path/cloud URI}``, where
            remote means the VM(s) on which this Task will eventually run on,
            and local means the node from which the task is launched.

        Returns:
          self: the current task, with file mounts updated.

        Raises:
          ValueError: if input paths are invalid.
        """
        if self.file_mounts is None:
            self.file_mounts = {}
        assert self.file_mounts is not None
        self.file_mounts.update(file_mounts)
        # For validation logic:
        return self.set_file_mounts(self.file_mounts)

    def set_storage_mounts(
        self,
        storage_mounts: Optional[Dict[str, storage_lib.Storage]],
    ) -> 'Task':
        """Sets the storage mounts for this task.

        Storage mounts are a dictionary: ``{mount_path: sky.Storage object}``,
        each of which mounts a sky.Storage object (a cloud object store bucket)
        to a path inside the remote cluster.

        A sky.Storage object can be created by uploading from a local directory
        (setting ``source``), or backed by an existing cloud bucket (setting
        ``name`` to the bucket name; or setting ``source`` to the bucket URI).

        Example:
            .. code-block:: python

                task.set_storage_mounts({
                    '/remote/imagenet/': sky.Storage(name='my-bucket',
                                                     source='/local/imagenet'),
                })

        Args:
          storage_mounts: an optional dict of ``{mount_path: sky.Storage
            object}``, where mount_path is the path inside the remote VM(s)
            where the Storage object will be mounted on.

        Returns:
          self: The current task, with storage mounts set.

        Raises:
          ValueError: if input paths are invalid.
        """
        if storage_mounts is None:
            self.storage_mounts = {}
            # Clear the requires_fuse flag if no storage mounts are set.
            assert self.resources is not None, 'Task resources are required'
            return self
        for target, storage_obj in storage_mounts.items():
            # TODO(zhwu): /home/username/sky_workdir as the target path need
            # to be filtered out as well.
            if (
                target == constants.KONDUKTOR_REMOTE_WORKDIR
                and self.workdir is not None
            ):
                with ux_utils.print_exception_no_traceback():
                    raise ValueError(
                        f'Cannot use {constants.KONDUKTOR_REMOTE_WORKDIR!r} as a '
                        'destination path of a file mount, as it will be used '
                        'by the workdir. If uploading a file/folder to the '
                        'workdir is needed, please specify the full path to '
                        'the file/folder.'
                    )

            if data_utils.is_cloud_store_url(target):
                with ux_utils.print_exception_no_traceback():
                    raise ValueError(
                        'Storage mount destination path cannot be cloud storage'
                    )

            assert (
                storage_obj.mode == storage_lib.StorageMode.COPY
            ), 'Only COPY mode is supported for storage mounts'
            # TODO(asaiacai): can decide if we want to just delete this
            # if storage_obj.mode == storage_lib.StorageMode.MOUNT:
            #     # If any storage is using MOUNT mode, we need to enable FUSE in
            #     # the resources.
            #     for r in self.resources:
            #         r.requires_fuse = True

        # Storage source validation is done in Storage object
        self.storage_mounts = storage_mounts
        return self

    def get_local_to_remote_file_mounts(self) -> Optional[Dict[str, str]]:
        """Returns file mounts of the form (dst=VM path, src=local path).

        Any cloud object store URIs (gs://, s3://, etc.), either as source or
        destination, are not included.

        INTERNAL: this method is internal-facing.
        """
        if self.file_mounts is None:
            return None
        d = {}
        for k, v in self.file_mounts.items():
            if not data_utils.is_cloud_store_url(
                k
            ) and not data_utils.is_cloud_store_url(v):
                d[k] = v
        return d

    def to_yaml_config(self) -> Dict[str, Any]:
        """Returns a yaml-style dict representation of the task.

        INTERNAL: this method is internal-facing.
        """
        config = {}

        def add_if_not_none(key, value, no_empty: bool = False):
            if no_empty and not value:
                return
            if value is not None:
                config[key] = value

        add_if_not_none('name', self.name)

        tmp_resource_config: Dict[str, Any] = {}
        assert self.resources is not None, 'Resources are not defined'
        tmp_resource_config = self.resources.to_yaml_config()

        add_if_not_none('resources', tmp_resource_config)

        add_if_not_none('num_nodes', self.num_nodes)

        add_if_not_none('run', self.run)
        add_if_not_none('workdir', self.workdir)
        add_if_not_none('envs', self.envs, no_empty=True)

        add_if_not_none('file_mounts', {})

        if self.file_mounts is not None:
            config['file_mounts'].update(self.file_mounts)

        if self.storage_mounts is not None:
            config['file_mounts'].update(
                {
                    mount_path: storage.to_yaml_config()
                    for mount_path, storage in self.storage_mounts.items()
                }
            )
        return config



KONDUKTOR_CONTROLLER_VERSION = '0.1.0'

HARDWARE_XID_ERRORS = set(
    (
        48,
        *range(56, 59),
        *range(62, 65),
        *range(68, 78),
        *range(79, 87),
        *range(88, 90),
        92,
        *range(94, 106),
        *range(110, 121),
        *range(122, 126),
    )
)

# The set of all SXid error ids that are known to be harmless.
# See D.4 of https://docs.nvidia.com/datacenter/tesla/pdf/fabric-manager-user-guide.pdf
ALLOWLISTED_NVSWITCH_SXID_ERRORS = set(
    (
        11012,
        11021,
        11022,
        11023,
        12021,
        12023,
        15008,
        15011,
        19049,
        19055,
        19057,
        19059,
        19062,
        19065,
        19068,
        19071,
        24001,
        24002,
        24003,
        22013,
    )
)


POD_LOG_ERROR_REGEXES = [
    # possibly indicates degraded nvidia-FM in bad state
    r'`invalid device ordinal`',
]

DMESG_ERROR_REGEXES = [
    r'`(?i)nvidia-peermem nv_get_p2p_free_callback:\d+ '
    r'ERROR detected invalid context, skipping further processing`',
    r'`(?i)NVRM: xid`',
    r'`(?i)SXid`',
]
